#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Pre-defined datasets."""

import types
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import torch

import cebra.data as cebra_data
import cebra.helper as cebra_helper
from cebra.data.datatypes import Offset


class TensorDataset(cebra_data.SingleSessionDataset):
    """Discrete and/or continuously indexed dataset based on torch/numpy arrays.

    If dealing with datasets sufficiently small to fit :py:func:`numpy.array` or :py:class:`torch.Tensor`, this
    dataset is sufficient---the sampling auxiliary variable should be specified with a dataloader.
    Based on whether `continuous` and/or `discrete` auxiliary variables are provided, this class
    can be used with the discrete, continuous and/or mixed data loader classes.

    Args:
        neural:
            Array of dtype ``float`` or float Tensor of shape ``(N, D)``, containing neural activity over time.
        continuous:
            Array of dtype ```float`` or float Tensor of shape ``(N, d)``, containing the continuous behavior
            variables over the same time dimension.
        discrete:
            Array of dtype ```int64`` or integer Tensor of shape ``(N, d)``, containing the discrete behavior
            variables over the same time dimension.

    Example:

        >>> import cebra.data
        >>> import torch
        >>> data = torch.randn((100, 30))
        >>> index1 = torch.randn((100, 2))
        >>> index2 = torch.randint(0,5,(100, ))
        >>> dataset = cebra.data.datasets.TensorDataset(data, continuous=index1, discrete=index2)

    """

    def __init__(self,
                 neural: Union[torch.Tensor, npt.NDArray],
                 continuous: Union[torch.Tensor, npt.NDArray] = None,
                 discrete: Union[torch.Tensor, npt.NDArray] = None,
                 offset: Offset = Offset(0, 1),
                 device: str = "cpu"):
        super().__init__(device=device)
        self.neural = self._to_tensor(neural, check_dtype="float").float()
        self.continuous = self._to_tensor(continuous, check_dtype="float")
        self.discrete = self._to_tensor(discrete, check_dtype="int")
        if self.continuous is None and self.discrete is None:
            raise ValueError(
                "You have to pass at least one of the arguments 'continuous' or 'discrete'."
            )
        self.offset = offset

    def _to_tensor(
            self,
            array: Union[torch.Tensor, npt.NDArray],
            check_dtype: Optional[Literal["int",
                                          "float"]] = None) -> torch.Tensor:
        """Convert :py:func:`numpy.array` to :py:class:`torch.Tensor` if necessary and check the dtype.

        Args:
            array: Array to check.
            check_dtype: If not `None`, list of dtypes to which the values in `array`
                must belong to. Defaults to None.

        Returns:
            The `array` as a :py:class:`torch.Tensor`.
        """
        if array is None:
            return None
        if isinstance(array, np.ndarray):
            array = torch.from_numpy(array)
        if check_dtype is not None:
            if check_dtype not in ["int", "float"]:
                raise ValueError(
                    f"check_dtype must be 'int' or 'float', got {check_dtype}")
            if (check_dtype == "int" and not cebra_helper._is_integer(array)
               ) or (check_dtype == "float" and
                     not cebra_helper._is_floating(array)):
                raise TypeError(
                    f"Array has type {array.dtype} instead of {check_dtype}.")
        if cebra_helper._is_floating(array):
            array = array.float()
        if cebra_helper._is_integer(array):
            # NOTE(stes): Required for standardizing number format on
            # windows machines.
            array = array.long()
        return array

    @property
    def input_dimension(self) -> int:
        return self.neural.shape[1]

    @property
    def continuous_index(self):
        if self.continuous is None:
            raise NotImplementedError()
        return self.continuous

    @property
    def discrete_index(self):
        if self.discrete is None:
            raise NotImplementedError()
        return self.discrete

    def __len__(self):
        return len(self.neural)

    def __getitem__(self, index):
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)


def _assert_datasets_same_device(
        datasets: List[cebra_data.SingleSessionDataset]) -> str:
    """Checks if the list of datasets are all on the same device.

    Args:
        datasets: List of datasets.

    Returns:
        The device name if all datasets are on the same device.

    Raises:
        ValueError: If datasets are not all on the same device.
    """
    devices = set([dataset.device for dataset in datasets])
    if len(devices) != 1:
        raise ValueError("Datasets are not all on the same device")
    return devices.pop()


class DatasetCollection(cebra_data.MultiSessionDataset):
    """Multi session dataset made up of a list of datasets.

    Args:
        *datasets: Collection of datasets to add to the collection. The order
            will be maintained for indexing.

    Example:

        >>> import cebra.data
        >>> import torch
        >>> session1 = torch.randn((100, 30))
        >>> session2 = torch.randn((100, 50))
        >>> index1 = torch.randn((100, 4))
        >>> index2 = torch.randn((100, 4)) # same index dim as index1
        >>> dataset = cebra.data.DatasetCollection(
        ...               cebra.data.TensorDataset(session1, continuous=index1),
        ...               cebra.data.TensorDataset(session2, continuous=index2))

    """

    def _has_not_none_attribute(self, obj, key):
        """Check if obj.key exists.

        Returns:
            ``True`` if the key exists and is not ``None``. ``False`` if
            either the key does not exist, the attribute has value ``None``
            or the access raises a ``NotImplementedError``.
        """
        try:
            if hasattr(obj, key):
                if getattr(obj, key) is not None:
                    return True
        except NotImplementedError:
            return False
        return False

    def _unpack_dataset_arguments(
        self, datasets: Tuple[cebra_data.SingleSessionDataset]
    ) -> List[cebra_data.SingleSessionDataset]:
        if len(datasets) == 0:
            raise ValueError("Need to supply at least one dataset.")
        elif len(datasets) == 1:
            (dataset_generator,) = datasets
            if isinstance(dataset_generator, types.GeneratorType):
                return list(dataset_generator)
            else:
                raise ValueError(
                    "You need to specify either a single generator, "
                    "or multiple SingleSessionDataset instances.")
        else:
            return datasets

    def __init__(
        self,
        *datasets: cebra_data.SingleSessionDataset,
    ):
        self._datasets: List[
            cebra_data.SingleSessionDataset] = self._unpack_dataset_arguments(
                datasets)

        device = _assert_datasets_same_device(self._datasets)
        super().__init__(device=device)

        continuous = all(
            self._has_not_none_attribute(session, "continuous_index")
            for session in self.iter_sessions())
        discrete = all(
            self._has_not_none_attribute(session, "discrete_index")
            for session in self.iter_sessions())

        if not (continuous or discrete):
            raise ValueError(
                "The provided datasets need to define either continuous or discrete indices, "
                "or both. Continuous: {continuous}; discrete: {discrete}. "
                "Note that _all_ provided datasets need to define the indexing function of choice."
            )

        if continuous:
            self._cindex = torch.cat(list(
                self._iter_property("continuous_index")),
                                     dim=0)
        else:
            self._cindex = None
        if discrete:
            self._dindex = torch.cat(list(
                self._iter_property("discrete_index")),
                                     dim=0)
        else:
            self._dindex = None

    @property
    def num_sessions(self) -> int:
        """The number of sessions in the dataset."""
        return len(self._datasets)

    @property
    def input_dimension(self):
        return super().input_dimension

    def get_input_dimension(self, session_id: int) -> int:
        """Get the feature dimension of the required session.

        Args:
            session_id: The session ID, an integer between 0 and
                :py:attr:`num_sessions`.

        Returns:
            A single session input dimension for the requested session id.
        """
        return self.get_session(session_id).input_dimension

    def get_session(self, session_id: int) -> cebra_data.SingleSessionDataset:
        """Get the dataset for the specified session.

        Args:
            session_id: The session ID, an integer between 0 and
                :py:attr:`num_sessions`.

        Returns:
            A single session dataset for the requested session
            id.
        """
        return self._datasets[session_id]

    @property
    def continuous_index(self) -> torch.Tensor:
        return self._cindex

    @property
    def discrete_index(self) -> torch.Tensor:
        return self._dindex

    def _apply(self, func):
        return (func(data) for data in self.iter_sessions())

    def _iter_property(self, attr):
        return (getattr(data, attr) for data in self.iter_sessions())



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
import os

import joblib as jl

import cebra.datasets
from cebra.datasets import get_datapath
from cebra.datasets import monkey_reaching


def save_allen_decoding_dataset(savepath=get_datapath("allen_preload/")):
    """Load and save allen decoding dataset for Ca.
    Load and save all neural and behavioral data relevant for allen decoding dataset to reduce data loading time for the experiments using the shared data.
    It saves Ca data for the neuron numbers (10-1000), 5 different seeds for sampling the neurons and the train/test splits.
    Args:
        savepath: The directory to save the loaded data.
    """
    for n in [10, 30, 50, 100, 200, 400, 600, 800, 900, 1000]:
        for seed in [111, 222, 333, 444, 555]:
            for split_flag in ["train", "test"]:
                for modality in ["ca"]:
                    dataname = (
                        f"allen-movie1-{modality}-decoding-{n}-{split_flag}-{seed}"
                    )
                    data = cebra.datasets.init(dataname)
                    print(f"Initiated {dataname}")
                    jl.dump({"neural": data.neural},
                            f"{savepath}/{dataname}.jl")
                    print(f"{savepath}/{dataname}.jl")


def save_allen_dataset(savepath=get_datapath("allen_preload/")):
    """Load and save complete allen dataset for Ca.
    Load and save all neural and behavioral data relevant for allen decoding dataset to reduce data loading time for the experiments using the shared data.
    It saves Ca data for the neuron numbers (10-1000), 5 different seeds for sampling the neurons.
    Args:
        savepath: The directory to save the loaded data.
    """
    for n in [10, 30, 50, 100, 200, 400, 600, 800, 900, 1000]:
        for seed in [111, 222, 333, 444, 555]:
            for modality in ["neuropixel"]:
                dataname = f"allen-movie1-{modality}-{n}-{seed}"
                data = cebra.datasets.init(dataname)
                print(f"Initiated {dataname}")
                jl.dump({"neural": data.neural}, f"{savepath}/{dataname}.jl")
                print(f"{savepath}/{dataname}.jl")


def save_monkey_dataset(
        savepath: str = get_datapath("monkey_reaching_preload_smth_40/"),):
    """Load and save monkey reaching dataset.
    Load and save all neural and behavioral data relevant for monkey reaching dataset to reduce data loading time for the experiments using the shared data.
    It saves for all possible trials types ('active', 'passive', 'all') and the splits ('all', 'train', 'valid', 'test').
    Args:
        savepath: The directory to save the loaded data.
    """

    for session in ["active", "passive", "all"]:
        for split in ["all", "train", "valid", "test"]:
            data = monkey_reaching._load_data(session=session, split=split)
            dataname = f"{session}_{split}.jl"
            print(os.path.join(savepath, dataname))
            jl.dump(data, os.path.join(savepath, dataname))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--savepath", default=get_datapath(), type=str)
    args = parser.parse_args()
    save_allen_decoding_dataset(os.path.join(args.savepath, "allen_preload"))
    save_allen_dataset(os.path.join(args.savepath, "allen_preload"))
    save_monkey_dataset(os.path.join(args.savepath, "monkey_reaching_preload"))



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import _util
import torch
from torch import nn

import cebra.io


class HasDeviceDummy(cebra.io.HasDevice):

    def __init__(self):
        super().__init__()
        self.foo = torch.tensor([42])
        self.bar = torch.tensor([42])
        self.baz = nn.Parameter(torch.tensor([42.0]))


class Container(cebra.io.HasDevice):

    def __init__(self):
        super().__init__()
        self.foo = nn.Parameter(torch.tensor([42.0]))
        self.bar = nn.Linear(3, 3)
        self.baz = HasDeviceDummy()


class MoveToDeviceImplicit(cebra.io.HasDevice):

    def __init__(self):
        super().__init__()
        # sets the device to CPU
        self.cpu = torch.tensor([42], device="cpu")
        self.move_to_cpu = torch.tensor([42], device="cuda")
        assert self.move_to_cpu.device.type == "cpu"


class MoveToDeviceExplicit(cebra.io.HasDevice):

    def __init__(self):
        super().__init__("cuda")
        # sets the device to CPU
        self.move_to_cuda = torch.tensor([42], device="cpu")
        assert self.move_to_cuda.device.type == "cuda"


@_util.requires_cuda
def test_move_to_device_implicit():
    MoveToDeviceImplicit()


@_util.requires_cuda
def test_move_to_device_explicit():
    MoveToDeviceExplicit()


def _assert_device(obj, device):
    if isinstance(obj, nn.Module):
        for p in obj.parameters():
            assert p.device.type == device
    else:
        assert obj.device.type == device


@_util.requires_cuda
def test_has_device():
    dummy = HasDeviceDummy()
    assert dummy.device == "cpu"
    assert dummy._tensors == {"foo", "bar", "baz"}
    assert dummy.foo.device.type == "cpu"
    assert dummy.baz.device.type == "cpu"
    dummy.to("cuda")
    assert dummy.device == "cuda"
    assert dummy.foo.device.type == "cuda"
    assert dummy.baz.device.type == "cuda"


@_util.requires_cuda
def test_has_device_nested():
    container = Container()
    assert container.device == "cpu"
    container.to("cuda")
    assert container.baz.device == "cuda"
    assert container.bar.weight.device.type == "cuda"
    assert container.baz.foo.device.type == "cuda"
    assert container.baz.baz.device.type == "cuda"
    assert container.foo.device.type == "cuda"
    _assert_device(container.bar, "cuda")



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Continuous variable multi-session sampling."""

import numpy as np
import torch

import cebra.distributions as cebra_distr
import cebra.io

_device = cebra.io.device()


def _search(data, query):
    if query.ndim == 1:
        query = query[None]
    return np.argmin(abs(data[None, :] - query[:, None]).sum(-1), axis=1)


def _invert_index(idx):
    """Invert an indexing function

    Let the given array define a function v: [N]->[N], then
    the returned array defines its inverse.

    Example:

        >>> import numpy as np
        >>> idx = np.array([2,3,1,0])
        >>> idx_inv = _invert_index(idx)
        >>> print(idx[idx_inv].tolist())
        [0, 1, 2, 3]

    """
    out = np.zeros_like(idx)
    out[idx] = np.arange(len(idx))
    return out


class MultisessionSampler(cebra_distr.PriorDistribution,
                          cebra_distr.ConditionalDistribution):
    """Continuous multi-session sampling.

    Align embeddings across multiple sessions, using a continuous
    index. The transitions between index samples are computed across
    all sessions.

    Note:
        The batch dimension of positive samples are shuffled.
        Before applying the contrastive loss, either the reference samples
        need to be aligned with the positive samples, or vice versa:

        >>> import cebra.distributions.multisession as cebra_distributions_multisession
        >>> import cebra.integrations.sklearn.dataset as cebra_sklearn_dataset
        >>> import cebra.data
        >>> import torch
        >>> from torch import nn
        >>> # Multisession training: one model per dataset (different input dimensions)
        >>> session1 = torch.rand(100, 30)
        >>> session2 = torch.rand(100, 50)
        >>> index1 = torch.rand(100)
        >>> index2 = torch.rand(100)
        >>> num_features = 8
        >>> dataset = cebra.data.DatasetCollection(
        ...               cebra_sklearn_dataset.SklearnDataset(session1, (index1, )),
        ...               cebra_sklearn_dataset.SklearnDataset(session2, (index2, )))
        >>> model = nn.ModuleList([
        ...                cebra.models.init(
        ...                    name="offset1-model",
        ...                    num_neurons=dataset.input_dimension,
        ...                    num_units=32,
        ...                    num_output=num_features,
        ...                ) for dataset in dataset.iter_sessions()]).to("cpu")
        >>> sampler = cebra_distributions_multisession.MultisessionSampler(dataset, time_offset=10)

        >>> # ref and pos samples from all datasets
        >>> ref = sampler.sample_prior(100)
        >>> pos, idx, rev_idx = sampler.sample_conditional(ref)
        >>> ref = torch.LongTensor(ref)
        >>> pos = torch.LongTensor(pos)

        >>> # Then the embedding spaces can be concatenated
        >>> refs, poss = [], []
        >>> for i in range(len(model)):
        ...     refs.append(model[i](dataset._datasets[i][ref[i]]))
        ...     poss.append(model[i](dataset._datasets[i][pos[i]]))
        >>> ref = torch.stack(refs, dim=0)
        >>> pos = torch.stack(poss, dim=0)

        >>> # Now the index can be applied to the stacked features,
        >>> # to align reference to positive samples
        >>> aligned_ref = sampler.mix(ref, idx)
        >>> reference = aligned_ref.view(-1, num_features)
        >>> positive = pos.view(-1, num_features)
        >>> loss = (reference - positive)**2

        >>> # .. or the reverse index, to align positive to reference samples
        >>> aligned_pos = sampler.mix(pos, rev_idx)
        >>> reference = ref.view(-1, num_features)
        >>> positive = aligned_pos.view(-1, num_features)
        >>> loss = (ref - pos)**2

        The reason for this implementation is that ``dataset[i]`` will in
        general have different dimensions (for example, number of neurons),
        per session. In contrast to the reference and positive
        indices, this data cannot be stacked and models need to be applied
        session by session.

    After data processing, the dimensionality of the returned features
    matches. The resulting embeddings can be concatenated, and shuffling
    (across the session axis) can be applied to the reference samples, or
    reversed for the positive samples.

    Note:
        This function does currently not support explicitly selected
        discrete indices. They should be added as dimensions to the
        continuous index. More weight can be added to the discrete
        dimensions by using larger values in one-hot coding.

    TODO:
        * Add a dedicated sampler for mixed multi session sampling.
        * Add better CUDA support and refactor ``numpy`` implementation.
    """

    def __init__(self, dataset, time_offset):
        self.dataset = dataset

        # TODO(stes): implement in pytorch
        self.all_data = self.dataset.continuous_index.cpu().numpy()
        self.session_lengths = np.array(self.dataset.session_lengths)

        self.lengths = np.cumsum(self.session_lengths)
        self.lengths[1:] = self.lengths[:-1]
        self.lengths[0] = 0

        # TODO(stes): unify naming
        time_delta = time_offset
        self.time_difference = (torch.cat(
            [
                dataset.continuous_index[time_delta:] -
                dataset.continuous_index[:-time_delta]
                for dataset in self.dataset.iter_sessions()
            ],
            dim=0,
        ).cpu().numpy())

        self.index = [
            cebra_distr.ContinuousIndex(
                dataset.continuous_index.float().to(_device))
            for dataset in self.dataset.iter_sessions()
        ]

    @property
    def num_sessions(self) -> int:
        """The number of sessions in the index."""
        return len(self.lengths)

    def mix(self, array: np.ndarray, idx: np.ndarray):
        """Re-order array elements according to the given index mapping.

        The given array should be of the shape ``(session, batch, ...)`` and the
        indices should have length ``session x batch``, representing a mapping
        between indices.

        The resulting array will be rearranged such that
        ``out.reshape(session*batch, -1)[i] = array.reshape(session*batch, -1)[idx[i]]``

        For the inverse mapping, convert the indices first using ``_invert_index``
        function.

        Args:
            array: A 2D matrix containing samples for each session.
            idx: A list of indexes to re-order ``array`` on.
        """
        n, m = array.shape[:2]
        return array.reshape(n * m, -1)[idx].reshape(array.shape)

    def sample_prior(self, num_samples):
        # TODO(stes) implement empirical/uniform resampling
        ref_idx = np.random.uniform(0, 1, (self.num_sessions, num_samples))
        ref_idx = (ref_idx * self.session_lengths[:, None]).astype(int)
        return ref_idx

    def sample_conditional(self, idx: torch.Tensor) -> torch.Tensor:
        """Sample from the conditional distribution.

        Note:
            * Reference samples are sampled equally between sessions.
            * Queries are computed for each reference as in single-session,
              meaning by adding a random time shift to each reference sample.
            * In order to guarantee the same number of positive samples per
              session, queries are randomly assigned to a session and its
              corresponding positive sample is searched in that session only.
            * As a result, ref/pos pairing is shuffled and can be recovered
              the reverse shuffle operation.

        Args:
            idx: Reference indices, with dimension ``(session, batch)``.

        Returns:
            Positive indices (1st return value), which will be grouped by
            session and *not* match the reference indices.
            In addition, a mapping will be returned to apply the same shuffle operation
            that was applied to assign a query to a session along session/batch dimension
            to the reference indices (2nd return value), or reverse the shuffle operation
            (3rd return value).
            Returned shapes are ``(session, batch), (session, batch), (session, batch)``.

        TODO:
            * re-implement in pytorch for additional speed gains
        """

        shape = idx.shape
        # TODO(stes) unclear why we cannot restrict to 2d overall
        # idx has shape (2, #samples per batch)
        s = idx.shape[:2]
        idx = (idx + self.lengths[:, None]).flatten()

        diff_idx = torch.randint(len(self.time_difference), (len(idx),))
        query = self.all_data[idx] + self.time_difference[diff_idx]

        # shuffle operation to assign each query to a session
        idx = np.random.permutation(len(query))

        # TODO this part fails in Pytorch
        query = query[idx.reshape(s)]
        query = torch.from_numpy(query).to(_device)

        pos_idx = torch.zeros(shape, device=_device).long()
        for i in range(self.num_sessions):
            pos_idx[i] = self.index[i].search(query[i])
        pos_idx = pos_idx.cpu().numpy()

        # reverse indices to recover the ref/pos samples matching
        idx_rev = _invert_index(idx)
        return pos_idx, idx, idx_rev

    def __getitem__(self, pos_idx):
        pos_samples = np.zeros(pos_idx.shape[:2] + (self.data.shape[2],))
        for i in range(self.num_sessions):
            pos_samples[i] = self.data[i][pos_idx[i]]
        return pos_samples


class DiscreteMultisessionSampler(cebra_distr.PriorDistribution,
                                  cebra_distr.ConditionalDistribution):
    """Discrete multi-session sampling.

    Discrete indices don't need to be aligned. Positive pairs are found
    by matching the discrete index in randomly assigned sessions.

    After data processing, the dimensionality of the returned features
    matches. The resulting embeddings can be concatenated, and shuffling
    (across the session axis) can be applied to the reference samples, or
    reversed for the positive samples.

    TODO:
        * Add better CUDA support and refactor ``numpy`` implementation.
    """

    def __init__(self, dataset):
        self.dataset = dataset

        # TODO(stes): implement in pytorch
        self.all_data = self.dataset.discrete_index.cpu().numpy()
        self.session_lengths = np.array(self.dataset.session_lengths)

        self.lengths = np.cumsum(self.session_lengths)
        self.lengths[1:] = self.lengths[:-1]
        self.lengths[0] = 0

        self.index = [
            cebra_distr.DiscreteUniform(
                dataset.discrete_index.int().to(_device))
            for dataset in self.dataset.iter_sessions()
        ]

    @property
    def num_sessions(self) -> int:
        """The number of sessions in the index."""
        return len(self.lengths)

    def mix(self, array: np.ndarray, idx: np.ndarray):
        """Re-order array elements according to the given index mapping.

        The given array should be of the shape ``(session, batch, ...)`` and the
        indices should have length ``session x batch``, representing a mapping
        between indices.

        The resulting array will be rearranged such that
        ``out.reshape(session*batch, -1)[i] = array.reshape(session*batch, -1)[idx[i]]``

        For the inverse mapping, convert the indices first using ``_invert_index``
        function.

        Args:
            array: A 2D matrix containing samples for each session.
            idx: A list of indexes to re-order ``array`` on.
        """
        n, m = array.shape[:2]
        return array.reshape(n * m, -1)[idx].reshape(array.shape)

    def sample_prior(self, num_samples):
        # TODO(stes) implement empirical/uniform resampling
        ref_idx = np.random.uniform(0, 1, (self.num_sessions, num_samples))
        ref_idx = (ref_idx * self.session_lengths[:, None]).astype(int)
        return ref_idx

    def sample_conditional(self, idx: torch.Tensor) -> torch.Tensor:
        """Sample from the conditional distribution.

        Note:
            * Reference samples are sampled equally between sessions.
            * In order to guarantee the same number of positive samples per
              session, reference samples are randomly assigned to a session and its
              corresponding positive sample is searched in that session only.
            * As a result, ref/pos pairing is shuffled and can be recovered
              the reverse shuffle operation.

        Args:
            idx: Reference indices, with dimension ``(session, batch)``.

        Returns:
            Positive indices (1st return value), which will be grouped by
            session and *not* match the reference indices.
            In addition, a mapping will be returned to apply the same shuffle operation
            that was applied to assign reference samples to a session along session/batch dimension
            (2nd return value), or reverse the shuffle operation (3rd return value).
            Returned shapes are ``(session, batch), (session, batch), (session, batch)``.

        TODO:
            * re-implement in pytorch for additional speed gains
        """

        shape = idx.shape
        # TODO(stes) unclear why we cannot restrict to 2d overall
        # idx has shape (2, #samples per batch)
        s = idx.shape[:2]
        idx_all = (idx + self.lengths[:, None]).flatten()

        # get discrete indices
        query = self.all_data[idx_all]

        # shuffle operation to assign each index to a session
        idx = np.random.permutation(len(query))

        # TODO this part fails in Pytorch
        # apply shuffle
        query = query[idx.reshape(s)]
        query = torch.from_numpy(query).to(_device)

        # sample conditional for each assigned session
        pos_idx = torch.zeros(shape, device=_device).long()
        for i in range(self.num_sessions):
            pos_idx[i] = self.index[i].sample_conditional(query[i])
        pos_idx = pos_idx.cpu().numpy()

        # reverse indices to recover the ref/pos samples matching
        idx_rev = _invert_index(idx)
        return pos_idx, idx, idx_rev

    def __getitem__(self, pos_idx):
        pos_samples = np.zeros(pos_idx.shape[:2] + (self.data.shape[2],))
        for i in range(self.num_sessions):
            pos_samples[i] = self.data[i][pos_idx[i]]
        return pos_samples



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Allen pseudomouse Neuropixels decoding dataset with train/test split.

References:
    *https://allensdk.readthedocs.io/en/latest/visual_coding_neuropixels.html
    *Siegle, Joshua H., et al. "Survey of spiking in the mouse visual system reveals functional hierarchy." Nature 592.7852 (2021): 86-92.

"""
import pathlib

import joblib
import numpy as np
import torch
from numpy.random import Generator
from numpy.random import PCG64

import cebra.data
from cebra.datasets import get_datapath
from cebra.datasets import parametrize
from cebra.datasets.allen import ca_movie_decoding
from cebra.datasets.allen import NUM_NEURONS
from cebra.datasets.allen import SEEDS
from cebra.datasets.allen import SEEDS_DISJOINT

_DEFAULT_DATADIR = get_datapath()


@parametrize(
    "allen-movie-{num_movie}-neuropixel-{cortex}-{num_neurons}-{split_flag}-10-{seed}",
    num_movie=["one"],
    cortex=["VISp", "VISpm", "VISam", "VISrl", "VISal", "VISl"],
    num_neurons=NUM_NEURONS,
    split_flag=["train", "test"],
    seed=SEEDS,
)
class AllenNeuropixelMovieDecoding120HzCorticesDataset(
        ca_movie_decoding.AllenCaMoviesDataset,
        cebra.data.SingleSessionDataset):
    """A pseudomouse 120Hz Neuropixels dataset during the allen MOVIE1 stimulus with train/test split.

    A dataset of stacked 120HZ spike counts recorded in the visual cortices (VISp, VISpm, VISam, VISrl, VISal, VISl) of multiple mice
    during the first 10 repeats of the MOVIE1 stimulus in Brain Observatory 1.1 set.
    The units which ISI > 0.5, amplitude < 0.1, presence ratio < 0.95 are excluded.
    The continuous labels corresponding to a DINO embedding of each stimulus frame.
    The 10th repeat is held out as a test set and the remaining 9 repeats consist a train set.

    Args:
        cortext: The cortical area to sample the neurons from.
        num_neurons: The number of neurons to randomly sample from the stacked pseudomouse neurons. Choose from 10, 30, 50, 100, 200, 400, 600, 800, 900, 1000.
        split_flag: The split to load. Choose between `train` and `test`.
        seed: The random seeds for sampling neurons.

    """

    def __init__(self, num_movie, cortex, num_neurons, split_flag, seed):
        super(ca_movie_decoding.AllenCaMoviesDataset, self).__init__()
        self.num_neurons = num_neurons
        self.seed = seed
        self.split_flag = split_flag
        frame_feature = self._get_video_features(num_movie)
        mice_data = self._get_pseudo_mice(cortex, num_movie)
        pseudo_mice = mice_data["neural"].T
        self.frames_index = mice_data["frames"]
        self.movie_len = int(pseudo_mice.shape[1])
        self.neurons_indices = self._sample_neurons(pseudo_mice)
        self._split(pseudo_mice, frame_feature)

    def _get_pseudo_mice(self, cortex: str, num_movie: str = "one"):
        """Load the pseudomice neuropixels data of the specified cortical area.

        Args:
            cortex: The visual cortical area.
        """
        path = pathlib.Path(
            _DEFAULT_DATADIR
        ) / "allen" / "allen_movie1_neuropixel" / cortex / "neuropixel_pseudomouse_120_filtered.jl"
        data = joblib.load(path)
        return data

    def _split(self, pseudo_mice, frame_feature):
        """Split the dataset into train and test set.

        The first 9 repeats are the train set and the last repeat of the stimulu block is the test set.

        Args:
            pseudo_mice: The pseudomouse neural data.
            frame_feature: The frame feature used as the behavior label.

        """

        if self.split_flag == "train":
            self.neural = (torch.from_numpy(
                pseudo_mice[self.neurons_indices, :int(self.movie_len / 10 *
                                                       9)]).float().T)
            self.index = frame_feature[self.frames_index[:int(self.movie_len /
                                                              10 * 9)]]
            self.frames_index = self.frames_index[:int(self.movie_len / 10 * 9)]
        elif self.split_flag == "test":
            self.neural = (torch.from_numpy(
                pseudo_mice[self.neurons_indices,
                            int(self.movie_len / 10 * 9):]).float().T)
            self.index = frame_feature[self.frames_index[int(self.movie_len /
                                                             10 * 9):]]
            self.frames_index = self.frames_index[int(self.movie_len / 10 * 9):]


@parametrize(
    "allen-movie-one-neuropixel-{cortex}-disjoint-{group}-{num_neurons}-{split_flag}-10-{seed}",
    cortex=["VISp", "VISam", "VISrl", "VISal"],
    num_neurons=[400],
    split_flag=["train", "test"],
    seed=SEEDS_DISJOINT,
    group=[0, 1],
)
class AllenNeuropixelMovie120HzCorticesDisjointDataset(
        AllenNeuropixelMovieDecoding120HzCorticesDataset):
    """A disjoint pseudomouse 120Hz Neuropixels dataset of  during the allen MOVIE1 stimulus with train/test splits.

    A dataset of stacked 120Hz spike counts recorded in the visual cortices (VISp, VISpm, VISam, VISrl, VISal, VISl) of multiple mice
    during the first 10 repeats of the MOVIE1 stimulus in Brain Observatory 1.1 set.
    The units which ISI > 0.5, amplitude < 0.1, presence ratio < 0.95 are excluded.
    The continuous labels corresponding to a DINO embedding of each stimulus frame.
    The disjoint sets of neurons are configured. For example, for each seed, group1 and group2 (called by `group` parameter) are disjoint to each other.
    The 10th repeat is held-out as a test set and the remaining 9 repeats consists a train set.

    Args:
        cortex: The cortical area to sample the neurons from.
        split_flag: The split to load. Choose between `train` and `test`.
        seed: The random seeds for sampling neurons.
        group: The index of the group among disjoint sets of the sampled neurons.

    """

    def __init__(self,
                 group,
                 num_neurons,
                 seed=111,
                 cortex="VISp",
                 split_flag="train",
                 frame_feature_path=pathlib.Path(_DEFAULT_DATADIR) / "allen" /
                 "features" / "allen_movies" / "vit_base" / "8" /
                 "movie_one_image_stack.npz" / "testfeat.pth"):
        self.split_flag = split_flag
        self.seed = seed
        self.group = group
        self.num_neurons = num_neurons
        data = joblib.load(
            pathlib.Path(_DEFAULT_DATADIR) / "allen" /
            "allen_movie1_neuropixel" / cortex /
            "neuropixel_pseudomouse_120_filtered.jl")
        pseudo_mice = data["neural"].T
        self.neurons_indices = self._sample_neurons(pseudo_mice)
        self.movie_len = pseudo_mice.shape[1]
        frame_feature = torch.load(frame_feature_path)
        self.frames_index = data["frames"]
        self._split(pseudo_mice, frame_feature)

    def _sample_neurons(self, pseudo_mice):
        """Randomly sample disjoint neurons.

        The sampled two groups of 400 neurons are non-overlapping.

        Args:
            pseudo_mice: The pseudomouse dataset.

        """

        sampler = Generator(PCG64(self.seed))
        permuted_neurons = sampler.permutation(pseudo_mice.shape[0])
        return np.array_split(permuted_neurons,
                              2)[self.group][:self.num_neurons]



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Generate synthetic datasets for benchmarking embedding quality.

References:
    Adapted from pi-VAE: https://github.com/zhd96/pi-vae/blob/main/code/pi_vae.py
"""
import argparse
import pathlib

import joblib as jl
import keras
import numpy as np
import poisson as poisson_utils
import scipy.stats
import tensorflow as tf

import third_party.pivae.pivae_code.pi_vae as pi_vae


def simulate_cont_data_diff_var(length: int, n_dim: int, noise_func: str):
    """Generate a synthetic dataset with the chosen generative process.

    Adapted from pi-VAE; https://github.com/zhd96/pi-vae/blob/main/code/pi_vae.py.
    The 1D continuous label u is sampled from uniform distribution defined in [0, 2*pi].
    The corresponding 2D latent z is sampled from a normal distribution defined by the mean (u, 2*sin(u)) and the variance (0.6-0.3*|sin(u)|, 0.3*|sin(u)|).
    The `n_dim` firing rate is generated by feeding z into bijective mixing function modeled by RealNVP.
    Finally, observation X is generated by adding noise of the choice.

    Args:
        length: The length of the simulation or the number of the simulated samples.
        n_dim: The dimension of the observation.
        noise_func: The distribution used for generative process
    """
    ## simulate 2d z
    np.random.seed(777)

    u_true = np.random.uniform(0, 2 * np.pi, size=[length, 1])
    mu_true = np.hstack((u_true, 2 * np.sin(u_true)))
    var_true = 0.15 * np.abs(mu_true)
    var_true[:, 0] = 0.6 - var_true[:, 1]
    z_true = np.random.normal(0, 1, size=[length, 2
                                         ]) * np.sqrt(var_true) + mu_true
    z_true = np.hstack((z_true, np.zeros((z_true.shape[0], n_dim - 2))))

    ## simulate mean
    dim_x = z_true.shape[-1]
    permute_ind = []
    n_blk = 4
    for ii in range(n_blk):
        np.random.seed(ii)
        permute_ind.append(tf.convert_to_tensor(np.random.permutation(dim_x)))

    x_input = keras.layers.Input(shape=(dim_x,))
    x_output = pi_vae.realnvp_block(x_input)
    for ii in range(n_blk - 1):
        x_output = keras.layers.core.Lambda(pi_vae.perm_func,
                                            arguments={"ind": permute_ind[ii]
                                                      })(x_output)
        x_output = pi_vae.realnvp_block(x_output)

    realnvp_model = keras.models.Model(inputs=[x_input], outputs=x_output)
    mean_true = realnvp_model.predict(z_true)
    lam_true = np.exp(2.2 * np.tanh(mean_true))
    if noise_func is not None:
        x = noise_func(lam_true)
        return z_true, u_true, mean_true, lam_true, x
    else:
        return z_true, u_true, mean_true, lam_true


__noises = {}


def _register_noise(func):
    __noises[func.__name__] = func
    return func


@_register_noise
def poisson(x):
    """Apply poisson noise to the input.

    This setup corresponds to the synthetic dataset originally
    considered by Zhou and Wei (NeurIPS 2021) for pi-VAE benchmarking.

    Args:
        x: The rate parameter

    Returns:
        Samples with the specified rate, of same shape as the input.
    """
    return np.random.poisson(x)


@_register_noise
def gaussian(x):
    """Apply truncated Gaussian noise with unit variance to the input.

    Args:
        x: The mean

    Returns:
        The samples, of same shape as the input.
    """
    return scipy.stats.truncnorm.rvs(0, 1000, loc=x)


@_register_noise
def laplace(x):
    """Apply (post-truncated) Laplace noise to the input.

    Args:
        x: The mean

    Returns:
        The samples, of same shape as the input.
    """
    return np.clip(np.random.laplace(x), a_min=0, a_max=1000)


@_register_noise
def uniform(x):
    """Apply uniform noise from [0, 2) to the input.

    Args:
        x: The offset added to the noise samples.

    Returns:
        The samples, of same shape as the input.
    """
    return np.random.uniform(0, 2, x.shape) + x


@_register_noise
def t(x):
    """Apply student-t distributed noise to the input.

    Args:
        x: The mean

    Returns:
        The samples, of same shape as the input.
    """
    return scipy.stats.distributions.t.rvs(2, loc=x)


@_register_noise
def refractory_poisson(x):
    """TODO. Not implemented yet."""
    raise NotImplementedError()


if __name__ == "__main__":
    """Generate synthetic datasets with poisson, gaussian, laplace, uniform, t noise during generative process."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--save-path",
        type=str,
        default="/data/synthetic/",
        help="Directory to save the generated dataset.",
    )
    parser.add_argument(
        "--n-samples",
        type=int,
        default=15000,
        help="The number of the trials to generate",
    )
    parser.add_argument("--neurons",
                        type=int,
                        default=100,
                        help="The number of the neurons")
    parser.add_argument(
        "--noise",
        type=str,
        choices=list(__noises.keys()),
        help="The type of noise to add in the generative process",
    )
    parser.add_argument(
        "--scale",
        type=float,
        default=50,
        help=
        "The scaling factor to firing rate for generating poisson neurons with refractory period",
    )
    parser.add_argument(
        "--time-interval",
        type=float,
        default=3,
        help=
        "The time interval (sec) to sample spikes for generating poisson neurons with refractory period",
    )
    parser.add_argument(
        "--refractory-period",
        type=float,
        default=0.01,
        help="The refaractory period (sec) of neurons",
    )

    args = parser.parse_args()

    if args.noise != "refractory_poisson":
        func = __noises[args.noise]
        z_true, u_true, mean_true, lam_true, x = simulate_cont_data_diff_var(
            args.n_samples, args.neurons, func)
    else:
        z_true, u_true, mean_true, lam_true = simulate_cont_data_diff_var(
            args.n_samples, args.neurons, None)
        flattened_lam = lam_true.flatten()
        x = np.zeros_like(flattened_lam)
        for i, rate in enumerate(flattened_lam):
            neuron = poisson_utils.PoissonNeuron(
                spike_rate=rate * args.scale,
                num_repeats=1,
                time_interval=args.time_interval,
            )
            count = neuron._get_counts(refractory_period=args.refractory_period)
            x[i] = count
        x = x.reshape(lam_true.shape)

    jl.dump(
        {
            "z": z_true,
            "u": u_true,
            "lam": lam_true,
            "x": x
        },
        pathlib.Path(args.save_path) / f"continuous_label_{args.noise}.jl",
    )



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pathlib
import tempfile
from unittest.mock import patch

import numpy as np
import pytest
import requests
import torch

import cebra.data
import cebra.data.assets as cebra_data_assets
import cebra.datasets
import cebra.registry
from cebra.datasets import poisson

_DEFAULT_DATADIR = cebra.datasets.get_datapath()


def test_registry():
    """Check the registry: Are all functions defined and is the
    docstring correctly adapted?"""
    assert cebra.registry.is_registry(cebra.datasets)
    assert cebra.registry.is_registry(cebra.datasets, check_docs=True)


def test_factory():
    """Register a new dataset"""
    import cebra.datasets

    @cebra.datasets.register("test-data")
    class TestDataset:
        pass

    assert "test-data" in cebra.datasets.get_options()
    instance = cebra.datasets.init("test-data")
    assert isinstance(instance, TestDataset)


def test_demo():
    dataset = cebra.datasets.init("demo-discrete")
    indices = torch.arange(0, 5)
    batch = dataset[indices]

    assert len(batch) == len(indices)


@pytest.mark.requires_dataset
def test_hippocampus():
    pytest.skip("Outdated")
    dataset = cebra.datasets.init("rat-hippocampus-single")
    loader = cebra.data.ContinuousDataLoader(
        dataset=dataset,
        num_steps=10,
        batch_size=8,
        conditional="time_delta",
    )
    for batch in loader:
        assert (len(batch.reference) == len(batch.positive)) and (len(
            batch.positive) == len(batch.negative))
        break

    dataset = cebra.datasets.init("rats-hippocampus-multisubjects")
    loader = cebra.data.ContinuousMultiSessionDataLoader(
        dataset=dataset,
        num_steps=10,
        batch_size=8,
        conditional="time_delta",
    )
    for batch in loader:
        for b in batch:
            assert (len(b.reference) == len(b.positive)) and (len(
                b.positive) == len(b.negative))
        break


@pytest.mark.requires_dataset
def test_monkey():
    dataset = cebra.datasets.init(
        "area2-bump-pos-active-passive",
        path=pathlib.Path(_DEFAULT_DATADIR) / "monkey_reaching_preload_smth_40",
    )
    indices = torch.randint(0, len(dataset), (10,))
    assert len(indices) == len(dataset[indices])


@pytest.mark.requires_dataset
def test_allen():
    pytest.skip("Test takes too long")

    ca_dataset = cebra.datasets.init("allen-movie-one-ca-VISp-100-train-10-111")
    ca_loader = cebra.data.ContinuousDataLoader(
        dataset=ca_dataset,
        num_steps=10,
        batch_size=8,
        conditional="time_delta",
    )
    for batch in ca_loader:
        assert (len(batch.reference) == len(batch.positive)) and (len(
            batch.positive) == len(batch.negative))
        break
    joint_dataset = cebra.datasets.init(
        "allen-movie-one-ca-neuropixel-VISp-100-train-10-111")
    joint_loader = cebra.data.ContinuousMultiSessionDataLoader(
        dataset=joint_dataset,
        num_steps=10,
        batch_size=8,
        conditional="time_delta",
    )
    for batch in joint_loader:
        for b in batch:
            assert (len(b.reference) == len(b.positive)) and (len(
                b.positive) == len(b.negative))
        break


try:
    options = cebra.datasets.get_options("*")
    multisubject_options = cebra.datasets.get_options(
        "allen-movie1-ca-multi-session-*")
    multisubject_options.extend(
        cebra.datasets.get_options(
            "rat-hippocampus-multisubjects-3fold-trial-split*"))
except:  # noqa: E722
    options = []


@pytest.mark.requires_dataset
@pytest.mark.parametrize("options",
                         cebra.datasets.get_options("*",
                                                    expand_parametrized=False))
def test_options(options):
    assert len(options) > 0
    assert len(multisubject_options) > 0


@pytest.mark.requires_dataset
@pytest.mark.parametrize("dataset", options)
def test_all(dataset):
    import cebra.datasets

    data = cebra.datasets.init(dataset)
    assert (data.continuous_index is not None) or (data.discrete_index
                                                   is not None)
    assert isinstance(data, cebra.data.base.Dataset)


@pytest.mark.requires_dataset
@pytest.mark.parametrize("dataset", multisubject_options)
def test_all_multisubject(dataset):
    # NOTE(stes) In theory a duplicate of test_all, but allows to quickly double check #611 won't re-appear
    # in the future. Will keep it in, but fine to remove at a later point once dataset tests are optimized.
    import cebra.datasets
    data = cebra.datasets.init(dataset)
    assert (data.continuous_index is not None) or (data.discrete_index
                                                   is not None)
    assert isinstance(data, cebra.data.base.Dataset)


@pytest.mark.requires_dataset
@pytest.mark.parametrize("dataset", [
    "allen-movie1-ca-multi-session-leave2out-repeat-0-train",
    "allen-movie1-ca-multi-session-decoding-repeat-1-test",
    "rat-hippocampus-multisubjects-3fold-trial-split",
    "rat-hippocampus-multisubjects-3fold-trial-split-0"
])
def test_compat_fix611(dataset):
    """Check that confirm the fix applied in internal PR #611

    The PR removed the explicit continuous and discrete args from the
    datasets used to parametrize this function. We manually check that
    the continuous index is available, and no discrete index is set.

    https://github.com/AdaptiveMotorControlLab/CEBRA-dev/pull/613
    """
    import cebra.datasets
    data = cebra.datasets.init(dataset)
    assert (data.continuous_index is not None)
    assert (data.discrete_index is None)
    assert isinstance(data, cebra.data.datasets.DatasetCollection)


def _assert_histograms_close(values, histogram):
    max_counts = max(max(values), len(histogram))

    value_mean = values.mean()
    histogram_mean = (histogram *
                      np.arange(len(histogram))).sum() / histogram.sum()
    assert np.isclose(value_mean, histogram_mean, rtol=0.05)

    value_histogram = np.bincount(values, minlength=max_counts)
    # NOTE(stes) normalize the histograms to be able to use the same atol values in the histogram
    # test below
    value_histogram = value_histogram / float(value_histogram.sum())
    histogram = histogram / float(histogram.sum())
    if len(histogram) < len(value_histogram):
        histogram = np.pad(
            histogram,
            pad_width=[(0, len(value_histogram) - len(histogram))],
            mode="constant",
            constant_values=(0, 0),
        )

    assert value_histogram.shape == histogram.shape
    # NOTE(stes) while the relative tolerance here is quite high (20%), this is a tradeoff vs. speed.
    # For lowering the tolerance, the number of samples drawn in the test methods needs to be increased.
    assert np.allclose(value_histogram, histogram, atol=0.05, rtol=0.25)


def test_poisson_reference_implementation():
    spike_rate = 40
    num_repeats = 500

    neuron_model = poisson.PoissonNeuron(
        spike_rate=spike_rate,
        num_repeats=num_repeats,
    )

    def _check_histogram(bins, hist):
        assert len(bins) == len(hist)
        assert (hist >= 0).all()

    bins, hist = neuron_model.sample_spikes()
    _check_histogram(bins, hist)
    assert hist.sum() == num_repeats

    bins, hist = neuron_model.sample_poisson_estimate()
    _check_histogram(bins, hist)

    bins, hist = neuron_model.sample_poisson()
    _check_histogram(bins, hist)


@pytest.mark.parametrize("spike_rate", 10**np.linspace(0, 2.0))
def test_homogeneous_poisson_sampling(spike_rate):
    torch.manual_seed(0)
    np.random.seed(0)
    spike_rates = spike_rate * torch.ones((10, 2000, 1))
    spike_counts = poisson._sample_batch(spike_rates)
    assert spike_counts.shape == spike_rates.shape

    neuron_model = poisson.PoissonNeuron(spike_rate=spike_rate,
                                         num_repeats=spike_counts.numel())
    _, reference_counts = neuron_model.sample_poisson(
        range_=(0, spike_counts.max() + 1))

    _assert_histograms_close(spike_counts.flatten().numpy(), reference_counts)


@pytest.mark.parametrize(
    "spike_rate,refractory_period",
    [[10, 0.02], [30, 0.01], [50, 0.0], [80, 0.1], [100, 0.01], [1, 0.001],
     [2, 0.0]],
)
def test_poisson_sampling(spike_rate, refractory_period):
    torch.manual_seed(0)
    np.random.seed(0)
    spike_rates = spike_rate * torch.ones((10, 2000, 1))
    spike_counts = poisson._sample_batch(spike_rates,
                                         refractory_period=refractory_period)
    transform = poisson.PoissonNeuronTransform(
        num_neurons=10, refractory_period=refractory_period)
    spike_counts = transform(spike_rates)
    assert spike_counts.shape == spike_rates.shape

    neuron_model = poisson.PoissonNeuron(spike_rate=spike_rate,
                                         num_repeats=spike_counts.numel())

    _, reference_counts = neuron_model.sample_spikes(
        refractory_period=refractory_period)

    _assert_histograms_close(spike_counts.flatten().numpy(), reference_counts)


def parametrize_data(function):
    return pytest.mark.parametrize("filename, url, expected_checksum", [
        ("achilles.jl",
         "https://figshare.com/ndownloader/files/40849463?private_link=9f91576cbbcc8b0d8828",
         "c52f9b55cbc23c66d57f3842214058b8"),
        ("buddy.jl",
         "https://figshare.com/ndownloader/files/40849460?private_link=9f91576cbbcc8b0d8828",
         "36341322907708c466871bf04bc133c2"),
        ("cicero.jl",
         "https://figshare.com/ndownloader/files/40849457?private_link=9f91576cbbcc8b0d8828",
         "a83b02dbdc884fdd7e53df362499d42f"),
        ("gatsby.jl",
         "https://figshare.com/ndownloader/files/40849454?private_link=9f91576cbbcc8b0d8828",
         "2b889da48178b3155011c12555342813"),
        ("all_all.jl",
         "https://figshare.com/ndownloader/files/41668764?private_link=6fa4ee74a8f465ec7914",
         "dea556301fa4fafa86e28cf8621cab5a"),
        ("active_all.jl",
         "https://figshare.com/ndownloader/files/41668776?private_link=6fa4ee74a8f465ec7914",
         "c626acea5062122f5a68ef18d3e45e51"),
        ("passive_all.jl",
         "https://figshare.com/ndownloader/files/41668758?private_link=6fa4ee74a8f465ec7914",
         "bbb1bc9d8eec583a46f6673470fc98ad"),
    ])(function)


@pytest.mark.requires_dataset
@parametrize_data
def test_download_file_successful_download(filename, url, expected_checksum):
    with tempfile.TemporaryDirectory() as temp_dir:
        cebra_data_assets.download_file_with_progress_bar(
            url=url,
            expected_checksum=expected_checksum,
            location=temp_dir,
            file_name=filename)

        downloaded_checksum = cebra_data_assets.calculate_checksum(
            os.path.join(temp_dir, filename))
        assert downloaded_checksum == expected_checksum


@parametrize_data
def test_download_file_wrong_checksum(filename, url, expected_checksum):
    wrong_checksum = ''.join(reversed(expected_checksum))
    with tempfile.TemporaryDirectory() as temp_dir:
        with pytest.raises(RuntimeError):
            cebra_data_assets.download_file_with_progress_bar(
                url=url,
                expected_checksum=wrong_checksum,
                location=temp_dir,
                file_name=filename,
                retry_count=2)


@parametrize_data
def test_download_file_wrong_url(filename, url, expected_checksum):
    wrong_url = "https://figshare.com/wrongurl"
    with tempfile.TemporaryDirectory() as temp_dir:
        with pytest.raises(requests.HTTPError):
            cebra_data_assets.download_file_with_progress_bar(
                url=wrong_url,
                expected_checksum=expected_checksum,
                location=temp_dir,
                file_name=filename)


@parametrize_data
def test_download_file_wrong_content_disposition(filename, url,
                                                 expected_checksum):
    with tempfile.TemporaryDirectory() as temp_dir:
        with patch("requests.get") as mock_get:
            mock_response = mock_get.return_value
            mock_response.status_code = 200
            mock_response.headers = {}

            with pytest.raises(ValueError):
                cebra_data_assets.download_file_with_progress_bar(
                    url=url,
                    expected_checksum=expected_checksum,
                    location=temp_dir,
                    file_name=filename)

            mock_response.headers = {"Content-Disposition": "invalid_header"}
            with pytest.raises(ValueError):
                cebra_data_assets.download_file_with_progress_bar(
                    url=url,
                    expected_checksum=expected_checksum,
                    location=temp_dir,
                    file_name=filename)


@pytest.mark.parametrize("neural, continuous, discrete", [
    (np.random.randn(100, 30), np.random.randn(
        100, 2), np.random.randint(0, 5, (100,))),
    (np.random.randn(50, 20), None, np.random.randint(0, 3, (50,))),
    (np.random.randn(200, 40), np.random.randn(200, 5), None),
])
def test_tensor_dataset_initialization(neural, continuous, discrete):
    dataset = cebra.data.datasets.TensorDataset(neural,
                                                continuous=continuous,
                                                discrete=discrete)
    assert dataset.neural.shape == neural.shape
    if continuous is not None:
        assert dataset.continuous.shape == continuous.shape
    if discrete is not None:
        assert dataset.discrete.shape == discrete.shape


def test_tensor_dataset_invalid_initialization():
    neural = np.random.randn(100, 30)
    with pytest.raises(ValueError):
        cebra.data.datasets.TensorDataset(neural)


@pytest.mark.parametrize("neural, continuous, discrete", [
    (np.random.randn(100, 30), np.random.randn(
        100, 2), np.random.randint(0, 5, (100,))),
    (np.random.randn(50, 20), None, np.random.randint(0, 3, (50,))),
    (np.random.randn(200, 40), np.random.randn(200, 5), None),
])
def test_tensor_dataset_length(neural, continuous, discrete):
    dataset = cebra.data.datasets.TensorDataset(neural,
                                                continuous=continuous,
                                                discrete=discrete)
    assert len(dataset) == len(neural)


@pytest.mark.parametrize("neural, continuous, discrete", [
    (np.random.randn(100, 30), np.random.randn(
        100, 2), np.random.randint(0, 5, (100,))),
    (np.random.randn(50, 20), None, np.random.randint(0, 3, (50,))),
    (np.random.randn(200, 40), np.random.randn(200, 5), None),
])
def test_tensor_dataset_getitem(neural, continuous, discrete):
    dataset = cebra.data.datasets.TensorDataset(neural,
                                                continuous=continuous,
                                                discrete=discrete)
    index = torch.randint(0, len(dataset), (10,))
    batch = dataset[index]
    assert batch.shape[0] == len(index)
    assert batch.shape[1] == neural.shape[1]


def test_tensor_dataset_invalid_discrete_type():
    neural = np.random.randn(100, 30)
    continuous = np.random.randn(100, 2)
    discrete = np.random.randn(100, 2)  # Invalid type: float instead of int
    with pytest.raises(TypeError):
        cebra.data.datasets.TensorDataset(neural,
                                          continuous=continuous,
                                          discrete=discrete)


@pytest.mark.parametrize("array, check_dtype, expected_dtype", [
    (np.random.randn(100, 30), "float", torch.float32),
    (np.random.randint(0, 5, (100, 30)), "int", torch.int64),
    (torch.randn(100, 30), "float", torch.float32),
    (torch.randint(0, 5, (100, 30)), "int", torch.int64),
    (None, None, None),
])
def test_to_tensor(array, check_dtype, expected_dtype):
    dataset = cebra.data.datasets.TensorDataset(np.random.randn(10, 2),
                                                continuous=np.random.randn(
                                                    10, 2))
    result = dataset._to_tensor(array, check_dtype=check_dtype)
    if array is None:
        assert result is None
    else:
        assert isinstance(result, torch.Tensor)
        assert result.dtype == expected_dtype


def test_to_tensor_invalid_dtype():
    dataset = cebra.data.datasets.TensorDataset(np.random.randn(10, 2),
                                                continuous=np.random.randn(
                                                    10, 2))
    array = np.random.randn(100, 30)
    with pytest.raises(TypeError):
        dataset._to_tensor(array, check_dtype="int")
    array = np.random.randint(0, 5, (100, 30))
    with pytest.raises(TypeError):
        dataset._to_tensor(array, check_dtype="float")


def test_to_tensor_invalid_check_dtype():
    dataset = cebra.data.datasets.TensorDataset(np.random.randn(10, 2),
                                                continuous=np.random.randn(
                                                    10, 2))
    array = np.random.randn(100, 30)
    with pytest.raises(ValueError,
                       match="check_dtype must be 'int' or 'float', got"):
        dataset._to_tensor(array, check_dtype="invalid_dtype")



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Neural network models and criterions for training CEBRA models."""
import abc

import torch
import torch.nn.functional as F
from torch import nn

import cebra.data
import cebra.data.datatypes
import cebra.models.layers as cebra_layers
from cebra.models import register


def _check_torch_version(raise_error=False):
    current_version = tuple(
        [int(i) for i in torch.__version__.split(".")[:2] if len(i) > 0])
    required_version = (1, 12)
    if current_version < required_version:
        if raise_error:
            raise ImportError(
                f"PyTorch < 1.12 is not supported for models using "
                f"Dropout1D, but got PyTorch={torch.__version__}.")
        else:
            return False
    return True


def _register_conditionally(*args, **kwargs):
    if _check_torch_version(raise_error=False):
        return register(*args, **kwargs)
    else:

        def do_nothing(cls):
            return cls

        return do_nothing


class Model(nn.Module):
    """Base model for CEBRA experiments.

    The model is a pytorch ``nn.Module``. Features can be computed by
    calling the ``forward()`` or ``__call__`` method. This class should not be
    directly instantiated, and instead used as the base class for CEBRA
    models.

    Args:
        num_input: The number of input dimensions. The tensor passed to
            the ``forward`` method will have shape ``(batch, num_input, in_time)``.
        num_output: The number of output dimensions. The tensor returned
            by the ``forward`` method will have shape ``(batch, num_output, out_time)``.
        offset: A specification of the offset to the left and right of
            the signal due to the network's receptive field. The offset specifies the
            relation between the input and output times, ``in_time - out_time = len(offset)``.

    Attributes:
        num_input: The input dimensionality (of the input signal). When calling
            ``forward``, this is the dimensionality expected for the input
            argument. In typical applications of CEBRA, the input dimension
            corresponds to the number of neurons for neural data analysis, number
            of keypoints for kinematik analysis, or can also be the dimension
            of a feature space in case preprocessing happened before feeding the
            data into the model.
        num_output: The output dimensionality (of the embedding space).
            This is the feature dimension of value returned by
            ``forward``. Note that for models using normalization,
            the output dimension should be at least 3D, and 2D without
            normalization to learn meaningful embeddings. The output
            dimensionality is typically smaller than :py:attr:`num_input`,
            but this is not enforced.
    """

    def __init__(
        self,
        *,
        num_input: int,
        num_output: int,
        offset: cebra.data.datatypes.Offset = None,
    ):
        super().__init__()
        if num_input < 1:
            raise ValueError(
                f"Input dimension needs to be at least 1, but got {num_input}.")
        if num_output < 1:
            raise ValueError(
                f"Output dimension needs to be at least 1, but got {num_output}."
            )
        self.num_input: int = num_input
        self.num_output: int = num_output

    @abc.abstractmethod
    def get_offset(self) -> cebra.data.datatypes.Offset:
        """Offset between input and output sequence caused by the receptive field.

        The offset specifies the relation between the length of the input and output
        time sequences. The output sequence is ``len(offset)`` steps shorter than the
        input sequence. For input sequences of shape ``(*, *, len(offset))``, the model
        should return an output sequence that drops the last dimension (which would be 1).

        Returns
            The offset of the network. See :py:class:`cebra.data.datatypes.Offset` for full
            documentation.
        """
        raise NotImplementedError()


class ConvolutionalModelMixin:
    """Mixin for models that support operating on a time-series.

    The input for convolutional models should be ``batch, dim, time``
    and the convolution will be applied across the last dimension.
    """

    pass


class ResampleModelMixin:
    """Mixin for models that re-sample the signal over time."""

    @property
    def resample_factor(self) -> float:
        """The factor by which the signal is downsampled."""
        return NotImplementedError()


class HasFeatureEncoder:
    """Networks with an explicitly defined feature encoder."""

    @property
    def feature_encoder(self) -> nn.Module:
        return self.net


class ClassifierModel(Model, HasFeatureEncoder):
    """Base model for classifiers.

    Adds an additional :py:attr:`classifier` layer to the model which is lazily
    initialized after calling :py:meth:`set_output_num`.

    Args:
        num_input: The number of input units
        num_output: The number of output units
        offset: The offset introduced by the model's receptive field

    Attributes:
        features_encoder: The feature encoder to map the input tensor (2d or 3d depending
            on the exact model implementation) into a feature space of same dimension
        classifier: Map from the feature space to class scores
    """

    def __init__(
        self,
        *,
        num_input: int,
        num_output: int,
        offset: cebra.data.datatypes.Offset = None,
    ):
        super().__init__(num_input=num_input, num_output=num_output)
        self.classifier: nn.Module = None

    @abc.abstractmethod
    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        raise NotImplementedError

    def set_output_num(self, label_num: int, override: bool = False):
        """Set the number of output classes.

        Args:
            label_num: The number of labels to be added to the classifier layer.
            override: If `True`, override an existing classifier layer. If you
                passed the parameters of this model to an optimizer, make sure
                to correctly handle the replacement of the classifier there.
        """
        if self.classifier is None or override:
            self.classifier = nn.Linear(self.num_output, label_num)
        else:
            raise RuntimeError("classifier is already initialized.")

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """See :py:class:`ClassifierModel`."""
        features = self.feature_encoder.forward(inputs)
        features = F.relu(features)
        prediction = self.classifier(features)
        return features, prediction


class _OffsetModel(Model, HasFeatureEncoder):

    def __init__(self,
                 *layers,
                 num_input=None,
                 num_output=None,
                 normalize=True):
        super().__init__(num_input=num_input, num_output=num_output)

        if normalize:
            layers += (cebra_layers._Norm(),)
        layers += (cebra_layers.Squeeze(),)
        self.net = nn.Sequential(*layers)
        # TODO(stes) can this layer be removed? it is already added to
        # the self.net
        self.normalize = normalize

    def forward(self, inp):
        """Compute the embedding given the input signal.

        Args:
            inp: The input tensor of shape `num_samples x self.num_input x time`

        Returns:
            The output tensor of shape `num_samples x self.num_output x (time - receptive field)`.

        Based on the parameters used for initializing, the output embedding
        is normalized to the hypersphere (`normalize = True`).
        """
        return self.net(inp)


class ParameterCountMixin:
    """Add a parameter counter to a torch.nn.Module."""

    @property
    def num_parameters(self) -> int:
        """Total parameter count of the model."""
        return sum(param.numel() for param in self.parameters())

    @property
    def num_trainable_parameters(self) -> int:
        """Number of trainable parameters."""
        return sum(
            param.numel() for param in self.parameters() if param.requires_grad)


@register("offset10-model")
class Offset10Model(_OffsetModel, ConvolutionalModelMixin):
    """CEBRA model with a 10 sample receptive field."""

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 1:
            raise ValueError(
                f"Hidden dimension needs to be at least 1, but got {num_units}."
            )
        super().__init__(
            nn.Conv1d(num_neurons, num_units, 2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 3),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(5, 5)


@register("offset10-model-mse")
class Offset10ModelMSE(Offset10Model):
    """Symmetric model with 10 sample receptive field, without normalization.

    Suitable for use with InfoNCE metrics for Euclidean space.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=False):
        super().__init__(num_neurons, num_units, num_output, normalize)


@register("offset5-model")
class Offset5Model(_OffsetModel, ConvolutionalModelMixin):
    """CEBRA model with a 5 sample receptive field and output normalization."""

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        super().__init__(
            nn.Conv1d(num_neurons, num_units, 2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 2),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(2, 3)


@register("offset1-model-mse")
class Offset0ModelMSE(_OffsetModel):
    """CEBRA model with a single sample receptive field, without output normalization."""

    def __init__(self, num_neurons, num_units, num_output, normalize=False):
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_output * 30,
            ),
            nn.GELU(),
            nn.Linear(num_output * 30, num_output * 30),
            nn.GELU(),
            nn.Linear(num_output * 30, num_output * 10),
            nn.GELU(),
            nn.Linear(int(num_output * 10), num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset1-model")
class Offset0Model(_OffsetModel):
    """CEBRA model with a single sample receptive field, with output normalization."""

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 2:
            raise ValueError(
                f"Number of hidden units needs to be at least 2, but got {num_units}."
            )
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            nn.Linear(num_units, num_units),
            nn.GELU(),
            nn.Linear(num_units, int(num_units // 2)),
            nn.GELU(),
            nn.Linear(int(num_units // 2), num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset1-model-v2")
class Offset0Modelv2(_OffsetModel):
    """CEBRA model with a single sample receptive field, with output normalization.

    This is a variant of :py:class:`Offset0Model`.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 2:
            raise ValueError(
                f"Number of hidden units needs to be at least 2, but got {num_units}."
            )
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            nn.Linear(num_units, num_units),
            nn.GELU(),
            nn.Linear(num_units, num_units),
            nn.GELU(),
            nn.Linear(num_units, num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset1-model-v3")
class Offset0Modelv3(_OffsetModel):
    """CEBRA model with a single sample receptive field, with output normalization.

    This is a variant of :py:class:`Offset0Model`.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 2:
            raise ValueError(
                f"Number of hidden units needs to be at least 2, but got {num_units}."
            )
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            nn.Linear(num_units, num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset1-model-v4")
class Offset0Modelv4(_OffsetModel):
    """CEBRA model with a single sample receptive field, with output normalization.

    This is a variant of :py:class:`Offset0Model`.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 2:
            raise ValueError(
                f"Number of hidden units needs to be at least 2, but got {num_units}."
            )
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units),
                               nn.GELU(),
                               crop=(0, None)),
            cebra_layers._Skip(nn.Linear(num_units, num_units),
                               nn.GELU(),
                               crop=(0, None)),
            cebra_layers._Skip(nn.Linear(num_units, num_units),
                               nn.GELU(),
                               crop=(0, None)),
            cebra_layers._Skip(nn.Linear(num_units, num_units),
                               nn.GELU(),
                               crop=(0, None)),
            nn.Linear(num_units, num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset1-model-v5")
class Offset0Modelv5(_OffsetModel):
    """CEBRA model with a single sample receptive field, with output normalization.

    This is a variant of :py:class:`Offset0Model`.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 2:
            raise ValueError(
                f"Number of hidden units needs to be at least 2, but got {num_units}."
            )
        super().__init__(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            cebra_layers._Skip(nn.Linear(num_units, num_units), crop=(0, None)),
            nn.GELU(),
            nn.Linear(num_units, num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("resample-model",
          deprecated=True)  # NOTE(stes) deprecated name for compatibility
@register("offset40-model-4x-subsample")
class ResampleModel(_OffsetModel, ConvolutionalModelMixin, ResampleModelMixin):
    """CEBRA model with 40 sample receptive field, output normalization and 4x subsampling."""

    ##120Hz
    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        super().__init__(
            cebra_layers._MeanAndConv(num_neurons, num_units, 4, stride=2),
            nn.Conv1d(num_neurons + num_units, num_units, 3, stride=2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 3),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    @property
    def resample_factor(self):
        return 4

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(20, 20)


@register("resample5-model", deprecated=True)
@register("offset20-model-4x-subsample")
class Resample5Model(_OffsetModel, ConvolutionalModelMixin, ResampleModelMixin):
    """CEBRA model with 20 sample receptive field, output normalization and 4x subsampling."""

    ##120Hz
    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        super().__init__(
            cebra_layers._MeanAndConv(num_neurons, num_units, 4, stride=2),
            nn.Conv1d(num_neurons + num_units, num_units, 3, stride=2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 2),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    @property
    def resample_factor(self):
        return 4

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(10, 10)


@register("resample1-model", deprecated=True)
@register("offset4-model-2x-subsample")
class Resample1Model(_OffsetModel, ResampleModelMixin):
    """CEBRA model with 4 sample receptive field, output normalization and 2x subsampling.

    This model is not convolutional, and needs to be applied to fixed ``(N, d, 4)`` inputs.
    """

    ##120Hz
    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        super().__init__(
            cebra_layers._MeanAndConv(num_neurons, num_units, 4, stride=2),
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons + num_units,
                num_units,
            ),
            nn.GELU(),
            nn.Linear(num_units, num_units),
            nn.GELU(),
            nn.Linear(num_units, int(num_units // 2)),
            nn.GELU(),
            nn.Linear(int(num_units // 2), num_output),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    @property
    def resample_factor(self):
        return 2

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(2, 2)


@register("supervised10-model")
class SupervisedNN10(ClassifierModel):
    """A supervised model with 10 sample receptive field."""

    def __init__(self, num_neurons, num_units, num_output):
        super(SupervisedNN10, self).__init__(num_input=num_neurons,
                                             num_output=num_output)

        self.net = nn.Sequential(
            nn.Conv1d(num_neurons, num_units, 2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 3),
            cebra_layers.Squeeze(),
        )
        self.num_output = num_output

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(5, 5)


@register("supervised1-model")
class SupervisedNN1(ClassifierModel):
    """A supervised model with single sample receptive field."""

    def __init__(self, num_neurons, num_units, num_output):
        super(SupervisedNN1, self).__init__(num_input=num_neurons,
                                            num_output=num_output)

        self.net = nn.Sequential(
            nn.Flatten(start_dim=1, end_dim=-1),
            nn.Linear(
                num_neurons,
                num_units,
            ),
            nn.GELU(),
            nn.Linear(num_units, num_units),
            nn.GELU(),
            nn.Linear(num_units, int(num_units // 2)),
            nn.GELU(),
            nn.Linear(int(num_units // 2), num_output),
            cebra_layers.Squeeze(),
        )
        self.num_output = num_output

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See :py:meth:`~.Model.get_offset`"""
        return cebra.data.Offset(0, 1)


@register("offset36-model")
class Offset36(_OffsetModel, ConvolutionalModelMixin):
    """CEBRA model with a 10 sample receptive field."""

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 1:
            raise ValueError(
                f"Hidden dimension needs to be at least 1, but got {num_units}."
            )
        super().__init__(
            nn.Conv1d(num_neurons, num_units, 2),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 3),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See `:py:meth:Model.get_offset`"""
        return cebra.data.Offset(18, 18)


@_register_conditionally("offset36-model-dropout")
class Offset36Dropout(_OffsetModel, ConvolutionalModelMixin):
    """CEBRA model with a 10 sample receptive field.

    Note:
        Requires ``torch>=1.12``.
    """

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 1:
            raise ValueError(
                f"Hidden dimension needs to be at least 1, but got {num_units}."
            )
        super().__init__(
            nn.Conv1d(num_neurons, num_units, 2),
            torch.nn.Dropout1d(p=0.1),
            nn.GELU(),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            cebra_layers._Skip(nn.Conv1d(num_units, num_units, 3), nn.GELU()),
            nn.Conv1d(num_units, num_output, 3),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See `:py:meth:Model.get_offset`"""
        return cebra.data.Offset(18, 18)


@_register_conditionally("offset36-model-more-dropout")
class Offset36Dropoutv2(_OffsetModel, ConvolutionalModelMixin):
    """CEBRA model with a 10 sample receptive field.

    Note:
        Requires ``torch>=1.12``.
    """

    def _make_layers(self, num_units, p, n):
        return [
            cebra_layers._Skip(torch.nn.Dropout1d(p=p),
                               nn.Conv1d(num_units, num_units, 3), nn.GELU())
            for _ in range(n)
        ]

    def __init__(self, num_neurons, num_units, num_output, normalize=True):
        if num_units < 1:
            raise ValueError(
                f"Hidden dimension needs to be at least 1, but got {num_units}."
            )
        super().__init__(
            nn.Conv1d(num_neurons, num_units, 2),
            torch.nn.Dropout1d(p=0.1),
            nn.GELU(),
            *self._make_layers(num_units, 0.1, 16),
            nn.Conv1d(num_units, num_output, 3),
            num_input=num_neurons,
            num_output=num_output,
            normalize=normalize,
        )

    def get_offset(self) -> cebra.data.datatypes.Offset:
        """See `:py:meth:Model.get_offset`"""
        return cebra.data.Offset(18, 18)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Single session solvers embed a single pair of time series."""

import copy

import literate_dataclasses as dataclasses
import torch

import cebra
import cebra.data
import cebra.models
import cebra.solver.base as abc_
from cebra.solver import register


@register("single-session")
class SingleSessionSolver(abc_.Solver):
    """Single session training with a symmetric encoder.

    This solver assumes that reference, positive and negative samples
    are processed by the same features encoder.
    """

    _variant_name = "single-session"

    def _inference(self, batch: cebra.data.Batch) -> cebra.data.Batch:
        """Given a batch of input examples, computes the feature representation/embedding.

        Args:
            batch: The input data, not necessarily aligned across the batch
                dimension. This means that ``batch.index`` specifies the map
                between reference/positive samples, if not equal ``None``.

        Returns:
            Processed batch of data. While the input data might not be aligned
            across the sample dimensions, the output data should be aligned and
            ``batch.index`` should be set to ``None``.
        """
        batch.to(self.device)
        ref = self.model(batch.reference)
        pos = self.model(batch.positive)
        neg = self.model(batch.negative)
        return cebra.data.Batch(ref, pos, neg)

    def get_embedding(self, data: torch.Tensor) -> torch.Tensor:
        """Return the embedding of the given input data.

        Note:
            This function assumes that the input data is sliced
            according to the receptive field of the model. The input data
            needs to match ``batch x dims x len(self.model.get_offset())``
            which is internally reduced to ``batch x dims x 1``. The last
            dimension is squeezed, and the output is of shape ``time x features``.

        This function does *not* perform checks for correctness of the
        input.

        Args:
            data: The input data tensor of shape `batch_time x dims x time`

        Returns:
            The output data tensor of shape `batch_time x features`.

        TODO:
            - Check if implementing checks in this function would downgrade
              speed during training/inference.
        """
        return self.model(data)[0].T


@register("single-session-aux")
@dataclasses.dataclass
class SingleSessionAuxVariableSolver(abc_.Solver):
    """Single session training for reference and positive/negative samples.

    This solver processes reference samples with a model different from
    processing the positive and
    negative samples. Requires that the ``reference_model`` is initialized
    to be different from the ``model`` used to process the positive and
    negative samples.

    Besides using an asymmetric encoder for the same modality, this solver
    also allows for e.g. time-contrastive learning across modalities, by
    using a reference model on modality A, and a different model processing
    the signal from modality B.
    """

    _variant_name = "single-session-aux"
    reference_model: torch.nn.Module = None

    def __post_init__(self):
        super().__post_init__()
        if self.reference_model is None:
            # NOTE(stes): This should work, according to this thread
            # https://discuss.pytorch.org/t/can-i-deepcopy-a-model/52192/19
            # and create a true copy of the model.
            self.reference_model = copy.deepcopy(self.model)
            self.reference_model.to(self.model.device)

    def _inference(self, batch):
        batch.to(self.device)
        ref = self.reference_model(batch.reference)
        pos = self.model(batch.positive)
        neg = self.model(batch.negative)
        return cebra.data.Batch(ref, pos, neg)


@register("single-session-hybrid")
@dataclasses.dataclass
class SingleSessionHybridSolver(abc_.MultiobjectiveSolver):
    """Single session training, contrasting neural data against behavior."""

    _variant_name = "single-session-hybrid"

    def _inference(self, batch: cebra.data.Batch) -> cebra.data.Batch:
        batch.to(self.device)
        behavior_ref = self.model(batch.reference)[0]
        behavior_pos = self.model(batch.positive[:int(len(batch.positive) //
                                                      2)])[0]
        behavior_neg = self.model(batch.negative)[0]
        time_pos = self.model(batch.positive[int(len(batch.positive) // 2):])[1]
        time_ref = self.model(batch.reference)[1]
        time_neg = self.model(batch.negative)[1]
        return cebra.data.Batch(behavior_ref, behavior_pos,
                                behavior_neg), cebra.data.Batch(
                                    time_ref, time_pos, time_neg)


@register("single-session-full")
@dataclasses.dataclass
class BatchSingleSessionSolver(SingleSessionSolver):
    """Optimize a model with batch gradient descent.

    Usage of this solver requires a sufficient amount of GPU memory. Using this solver
    is equivalent to using a single session solver with batch size set to dataset size,
    but requires less computation.
    """

    def fit(self, loader, *args, **kwargs):
        """TODO"""
        self.offset = loader.dataset.offset
        self.neural = loader.dataset.neural.T[None]
        if isinstance(self.model, cebra.models.ConvolutionalModelMixin):
            if self.offset is None:
                raise ValueError("Configure dataset, no offset found.")
            self._mode = "convolutional"
        else:
            self.offset = cebra.data.Offset(0, 1)
            self._mode = "fully_connected"
        super().fit(loader, *args, **kwargs)

    def get_embedding(self, data):
        """Compute the embedding of a full input dataset.

        For convolutional models that implement
        :py:class:`cebra.models.model.ConvolutionalModelMixin`),
        the embedding is computed via
        :py:meth:`.SingleSessionSolver.get_embedding`.

        For all other models, it is assumed that the data has shape
        ``(1, dim, time)`` and is transformed into ``(time, dim)``
        format.

        Args:
            data: The input data

        Returns:
            The output embedding of shape ``(time, dimension)``

        See Also:
            * :py:class:`cebra.models.model.ConvolutionalModelMixin`)
            * :py:meth:`.SingleSessionSolver.get_embedding`
        """
        if self._mode == "convolutional":
            return super().get_embedding(data)
        else:
            # data has shape (1, d, T)
            # output needs to be (T, d)
            return self.model(data[0].T)

    def _inference(self, batch: cebra.data.Batch) -> cebra.data.Batch:
        outputs = self.get_embedding(self.neural)
        idc = batch.positive - self.offset.left >= len(outputs)
        batch.positive[idc] = batch.reference[idc]

        return cebra.data.Batch(
            outputs[batch.reference - self.offset.left],
            outputs[batch.positive - self.offset.left],
            outputs[batch.negative - self.offset.left],
        )



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import torch

import cebra.data
import cebra.io


def parametrize_device(func):
    _devices = ("cpu", "cuda") if torch.cuda.is_available() else ("cpu",)
    return pytest.mark.parametrize("device", _devices)(func)


class LoadSpeed:

    def __init__(self, loader):
        self.loader = loader

    def __call__(self):
        n = 0
        for batch in self.loader:
            n += 1
        assert batch.reference.device.type == self.loader.device
        assert n == len(self.loader)


class RandomDataset(cebra.data.SingleSessionDataset):

    def __init__(self, N=100, d=5, device="cpu"):
        super().__init__(device=device)
        self._cindex = torch.randint(0, 5, (N, d), device=device).float()
        self._dindex = torch.randint(0, 5, (N,), device=device).long()
        self.neural = self._data = torch.randn((N, d), device=device)

    @property
    def input_dimension(self):
        return self._data.shape[1]

    def __len__(self):
        return len(self._data)

    @property
    def continuous_index(self):
        return self._cindex

    @property
    def discrete_index(self):
        return self._dindex

    def __getitem__(self, index):
        return self._data[index]


def test_offset():
    offset = cebra.data.Offset(5, 4)
    assert offset.left == 5
    assert offset.right == 4
    assert offset.left_slice == slice(0, 5)
    assert len(offset) == 5 + 4

    offset = cebra.data.Offset(0, 4)
    assert offset.left == 0
    assert offset.right == 4
    assert offset.left_slice == slice(0, 0)
    assert len(offset) == 4

    offset = cebra.data.Offset(5)
    assert offset.left == 5
    assert offset.right == 5
    assert offset.left_slice == slice(0, 5)
    assert offset.right_slice == slice(-5, None)
    assert len(offset) == 5 * 2

    with pytest.raises(ValueError, match="Invalid.*right"):
        offset = cebra.data.Offset(5, 0)
    with pytest.raises(ValueError, match="Invalid.*right"):
        offset = cebra.data.Offset(0, 0)
    with pytest.raises(ValueError, match="Invalid.*number"):
        offset = cebra.data.Offset(5, 5, 5)
    with pytest.raises(ValueError, match="Invalid.*bounds"):
        offset = cebra.data.Offset(-2, 4)
    with pytest.raises(ValueError, match="Invalid.*bounds"):
        offset = cebra.data.Offset(4, -2)


def _assert_dataset_on_correct_device(loader, device):
    assert hasattr(loader, "dataset")
    assert hasattr(loader, "device")
    assert isinstance(loader.dataset, cebra.io.HasDevice)
    assert loader.dataset.neural.device.type == device


def test_demo_data():
    if not torch.cuda.is_available():
        pytest.skip("Test only possible with CUDA.")

    dataset = RandomDataset(N=100, device="cuda")
    assert dataset.neural.device.type == "cuda"
    dataset.to("cpu")
    assert dataset.neural.device.type == "cpu"


def _assert_device(first, second):

    def _to_str(val):
        if isinstance(val, torch.device):
            return val.type
        return val

    assert _to_str(first) == _to_str(second)


@parametrize_device
@pytest.mark.parametrize(
    "data_name, loader_initfunc",
    [
        ("demo-discrete", cebra.data.DiscreteDataLoader),
        ("demo-continuous", cebra.data.ContinuousDataLoader),
        ("demo-mixed", cebra.data.MixedDataLoader),
    ],
)
def test_device(data_name, loader_initfunc, device):
    if not torch.cuda.is_available():
        pytest.skip("Test only possible with CUDA.")

    swap = {"cpu": "cuda", "cuda": "cpu"}
    other_device = swap.get(device)
    dataset = RandomDataset(N=100, device=other_device)

    loader = loader_initfunc(dataset, num_steps=10, batch_size=32)
    loader.to(device)
    assert loader.dataset == dataset
    _assert_device(loader.device, device)
    _assert_device(loader.dataset.device, device)

    _assert_device(loader.get_indices(10).reference.device, device)


@parametrize_device
@pytest.mark.parametrize("prior", ("uniform", "empirical"))
def test_discrete(prior, device, benchmark):
    dataset = RandomDataset(N=100, device=device)
    loader = cebra.data.DiscreteDataLoader(
        dataset=dataset,
        num_steps=10,
        batch_size=8,
        prior=prior,
    )
    _assert_dataset_on_correct_device(loader, device)
    load_speed = LoadSpeed(loader)
    benchmark(load_speed)


@parametrize_device
@pytest.mark.parametrize("conditional", ("time", "time_delta"))
def test_continuous(conditional, device, benchmark):
    dataset = RandomDataset(N=100, d=5, device=device)
    loader = cebra.data.ContinuousDataLoader(
        dataset=dataset,
        num_steps=10,
        batch_size=8,
        conditional=conditional,
    )
    _assert_dataset_on_correct_device(loader, device)
    load_speed = LoadSpeed(loader)
    benchmark(load_speed)


def _check_attributes(obj, is_list=False):
    if is_list:
        for obj_ in obj:
            _check_attributes(obj_, is_list=False)
    elif isinstance(obj, cebra.data.Batch) or isinstance(
            obj, cebra.data.BatchIndex):
        assert hasattr(obj, "positive")
        assert hasattr(obj, "negative")
        assert hasattr(obj, "reference")
    else:
        raise TypeError()


@parametrize_device
@pytest.mark.parametrize(
    "data_name, loader_initfunc",
    [
        ("demo-discrete", cebra.data.DiscreteDataLoader),
        ("demo-continuous", cebra.data.ContinuousDataLoader),
        ("demo-mixed", cebra.data.MixedDataLoader),
    ],
)
def test_singlesession_loader(data_name, loader_initfunc, device):
    data = cebra.datasets.init(data_name)
    data.to(device)
    loader = loader_initfunc(data, num_steps=10, batch_size=32)
    _assert_dataset_on_correct_device(loader, device)

    index = loader.get_indices(100)
    _check_attributes(index)

    for batch in loader:
        _check_attributes(batch)
        assert len(batch.positive) == 32


def test_multisession_cont_loader():
    data = cebra.datasets.MultiContinuous(nums_neural=[3, 4, 5],
                                          num_behavior=5,
                                          num_timepoints=100)
    loader = cebra.data.ContinuousMultiSessionDataLoader(
        data,
        num_steps=10,
        batch_size=32,
    )

    # Check the sampler
    assert hasattr(loader, "sampler")
    ref_idx = loader.sampler.sample_prior(1000)
    assert len(ref_idx) == 3  # num_sessions
    for session in range(3):
        assert ref_idx[session].max() < 100
    pos_idx, idx, idx_rev = loader.sampler.sample_conditional(ref_idx)

    assert pos_idx is not None
    assert idx is not None
    assert idx_rev is not None

    batch = next(iter(loader))

    def _mix(array, idx):
        shape = array.shape
        n, m = shape[:2]
        mixed = array.reshape(n * m, -1)[idx]
        print(mixed.shape, array.shape, idx.shape)
        return mixed.reshape(shape)

    def _process(batch, feature_dim=1):
        """Given list_i[(N,d_i)] batch, return (#session, N, feature_dim) tensor"""
        return torch.stack(
            [b.reference.flatten(1).mean(dim=1, keepdims=True) for b in batch],
            dim=0).repeat(1, 1, feature_dim)

    assert batch[0].reference.shape == (32, 3, 10)
    assert batch[1].reference.shape == (32, 4, 10)
    assert batch[2].reference.shape == (32, 5, 10)

    dummy_prediction = _process(batch, feature_dim=6)
    assert dummy_prediction.shape == (3, 32, 6)
    _mix(dummy_prediction, batch[0].index)


def test_multisession_disc_loader():
    data = cebra.datasets.MultiDiscrete(nums_neural=[3, 4, 5],
                                        num_timepoints=100)
    loader = cebra.data.DiscreteMultiSessionDataLoader(
        data,
        num_steps=10,
        batch_size=32,
    )

    # Check the sampler
    assert hasattr(loader, "sampler")
    ref_idx = loader.sampler.sample_prior(1000)
    assert len(ref_idx) == 3  # num_sessions

    # Check sample points are in session length range
    for session in range(3):
        assert ref_idx[session].max() < loader.sampler.session_lengths[session]
    pos_idx, idx, idx_rev = loader.sampler.sample_conditional(ref_idx)

    assert pos_idx is not None
    assert idx is not None
    assert idx_rev is not None

    batch = next(iter(loader))

    def _mix(array, idx):
        shape = array.shape
        n, m = shape[:2]
        mixed = array.reshape(n * m, -1)[idx]
        print(mixed.shape, array.shape, idx.shape)
        return mixed.reshape(shape)

    def _process(batch, feature_dim=1):
        """Given list_i[(N,d_i)] batch, return (#session, N, feature_dim) tensor"""
        return torch.stack(
            [b.reference.flatten(1).mean(dim=1, keepdims=True) for b in batch],
            dim=0).repeat(1, 1, feature_dim)

    assert batch[0].reference.shape == (32, 3, 10)
    assert batch[1].reference.shape == (32, 4, 10)
    assert batch[2].reference.shape == (32, 5, 10)

    dummy_prediction = _process(batch, feature_dim=6)
    assert dummy_prediction.shape == (3, 32, 6)
    _mix(dummy_prediction, batch[0].index)


@parametrize_device
@pytest.mark.parametrize(
    "data_name, loader_initfunc",
    [('demo-discrete-multisession', cebra.data.DiscreteMultiSessionDataLoader),
     ("demo-continuous-multisession",
      cebra.data.ContinuousMultiSessionDataLoader)],
)
def test_multisession_loader(data_name, loader_initfunc, device):
    # TODO change number of timepoints across the sessions

    data = cebra.datasets.init(data_name)
    kwargs = dict(num_steps=10, batch_size=32)
    loader = loader_initfunc(data, **kwargs)

    index = loader.get_indices(100)
    print(index[0])
    print(type(index))
    _check_attributes(index, is_list=False)

    for batch in loader:
        _check_attributes(batch, is_list=True)
        for session_batch in batch:
            assert len(session_batch.positive) == 32



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Rat hippocampus dataset

References:
    * Grosmark, A.D., and Buzsáki, G. (2016). Diversity in neural firing dynamics supports both rigid and learned
       hippocampal sequences. Science 351, 1440–1443.
    * Chen, Z., Grosmark, A.D., Penagos, H., and Wilson, M.A. (2016). Uncovering representations of sleep-associated
       hippocampal ensemble spike activity. Sci. Rep. 6, 32193.
    * Grosmark, A.D., Long J. and Buzsáki, G. (2016); Recordings from hippocampal area CA1, PRE, during and POST
      novel spatial learning. CRCNS.org. http://dx.doi.org/10.6080/K0862DC5

"""

import pathlib

import joblib
import numpy as np
import sklearn.model_selection
import sklearn.neighbors
import torch

import cebra.data
from cebra.datasets import get_datapath
from cebra.datasets import init
from cebra.datasets import parametrize
from cebra.datasets import register

_DEFAULT_DATADIR = get_datapath()

rat_dataset_urls = {
    "achilles": {
        "url":
            "https://figshare.com/ndownloader/files/40849463?private_link=9f91576cbbcc8b0d8828",
        "checksum":
            "c52f9b55cbc23c66d57f3842214058b8"
    },
    "buddy": {
        "url":
            "https://figshare.com/ndownloader/files/40849460?private_link=9f91576cbbcc8b0d8828",
        "checksum":
            "36341322907708c466871bf04bc133c2"
    },
    "cicero": {
        "url":
            "https://figshare.com/ndownloader/files/40849457?private_link=9f91576cbbcc8b0d8828",
        "checksum":
            "a83b02dbdc884fdd7e53df362499d42f"
    },
    "gatsby": {
        "url":
            "https://figshare.com/ndownloader/files/40849454?private_link=9f91576cbbcc8b0d8828",
        "checksum":
            "2b889da48178b3155011c12555342813"
    }
}


@register("rat-hippocampus-single")
@parametrize(
    "rat-hippocampus-single-{name}",
    name=["achilles", "buddy", "cicero", "gatsby"],
)
class SingleRatDataset(cebra.data.SingleSessionDataset):
    """A single rat hippocampus tetrode recording while the rat navigates on a linear track.

    Neural data is spike counts binned into 25ms time window and the continuous behavior label is position and the running direction (left, right) of a rat.
    The behavior label is structured as 3D array consists of position, right, and left.

    Args:
        name: The name of the rat to use. Choose among 'achilles', 'buddy', 'cicero' and 'gatsby'.

    """

    def __init__(self, name="achilles", root=_DEFAULT_DATADIR, download=True):
        location = pathlib.Path(root) / "rat_hippocampus"
        file_path = location / f"{name}.jl"

        super().__init__(download=download,
                         data_url=rat_dataset_urls[name]["url"],
                         data_checksum=rat_dataset_urls[name]["checksum"],
                         location=location,
                         file_name=f"{name}.jl")

        data = joblib.load(file_path)
        self.neural = torch.from_numpy(data["spikes"]).float()
        self.index = torch.from_numpy(data["position"]).float()
        self.name = name

    @property
    def input_dimension(self):
        return self.neural.size(1)

    @property
    def continuous_index(self):
        return self.index

    def __getitem__(self, index):
        """Return [ No.Samples x Neurons x 10 ]"""
        index = self.expand_index(index)
        return self.neural[index].transpose(2, 1)

    def __len__(self):
        return len(self.neural)

    def __repr__(self):
        return f"RatDataset(name: {self.name}, shape: {self.neural.shape})"

    def decode(self, x_train, y_train, x_test, y_test):
        """kNN decoding function.

        Perform a kNN decoding for n_neighbors = 1,4,9,26,25 with the given train set and test set.

        Args:
            x_train: The train set data
            y_train: The train set label
            x_test: The test set data
            y_test: The test set label

        """

        nn = np.power(np.linspace(1, 10, 6, dtype=int), 2)
        metric = {}
        for n in nn:
            knn = sklearn.neighbors.KNeighborsRegressor(n_neighbors=n)
            knn.fit(x_train, y_train)
            pred = knn.predict(x_test)
            err = np.median(abs(pred[:, 0] - y_test[:, 0]))
            score = knn.score(x_test, y_test)
            metric[f"n{n}_err"] = err
            metric[f"n{n}_r2"] = score
        return metric


@register("rat-hippocampus-3fold-trial-split")
@parametrize(
    "rat-hippocampus-{name}-3fold-trial-split-{split_no}",
    name=["achilles", "buddy", "cicero", "gatsby"],
    split_no=[0, 1, 2],
)
class SingleRatTrialSplitDataset(SingleRatDataset):
    """A single rat hippocampus tetrode recording while the rat navigates on a linear track with 3-fold splits.

    Neural data is spike counts binned into 25ms time window and the behavior is position and the running
    direction (left, right) of a rat. The behavior label is structured as 3D array consists of position,
    right, and left. The neural and behavior recordings are parsed into trials (a round trip from one end
    of the track) and the trials are split into a train, valid and test set with k=3 nested cross validation.

    Args:
        name: The name of a rat to use. Choose among 'achilles', 'buddy', 'cicero' and 'gatsby'.
        split_no: The `k` for k-fold split. Choose among 0, 1, 2.
        split: The split to use. Choose among 'train', 'valid', 'test', 'all', and 'wo_test'
            (all trials except test split).

    """

    def __init__(self,
                 name="achilles",
                 split_no=0,
                 split=None,
                 root=_DEFAULT_DATADIR):
        super().__init__(name=name, root=root)
        self.split_no = split_no
        self.split_name = split
        if split is not None:
            self._split(split)

    def _split(self, split, **kwargs):
        """Split the dataset into 3-fold nested cross validation scheme.

        The recordings are parsed into trials and split into a train, valid, test set with 3-fold nested cross validation scheme.

        Args:
            split: The split to use. Choose among 'train', 'valid', 'test', 'all', and 'wo_test'(all trials except test split).

        """

        direction_change_idx = np.where(self.index[1:, 1] != self.index[:-1,
                                                                        1])[0]
        trial_change_idx = np.append(
            np.insert(direction_change_idx[1::2], 0, 0), len(self.index))
        total_trials_num = len(trial_change_idx) - 1

        outer_folds = np.array_split(
            np.arange(total_trials_num),
            3)  ## Divide data into 3 equal trial-sized array
        inner_folds = sklearn.model_selection.KFold(n_splits=3,
                                                    random_state=None,
                                                    shuffle=False)
        ## in each outer fold array, make train, valid, test split

        train_trials = []
        valid_trials = []
        test_trials = []

        for out_fold in outer_folds:
            train_trial, val_test_trial = list(
                inner_folds.split(out_fold))[self.split_no]
            test_trial, valid_trial = np.array_split(val_test_trial, 2)
            train_trials.extend(np.array(out_fold)[train_trial])
            valid_trials.extend(np.array(out_fold)[valid_trial])
            test_trials.extend(np.array(out_fold)[test_trial])

        if split == "train":
            trials = train_trials
        elif split == "valid":
            trials = valid_trials
        elif split == "test":
            trials = test_trials
        elif split == "all":
            trials = np.arange(total_trials_num)
        elif split == "wo_test":
            trials = np.concatenate([train_trials, valid_trials])
        else:
            raise ValueError(
                f"'{split}' is not a valid split. Use 'train', 'valid' or 'test'"
            )
        self.selected_indices = tuple(
            slice(trial_change_idx[i], trial_change_idx[i + 1]) for i in trials)
        self.neural = torch.cat([
            self.neural[trial_change_idx[i]:trial_change_idx[i + 1]]
            for i in trials
        ])
        self.index = torch.cat([
            self.index[trial_change_idx[i]:trial_change_idx[i + 1]]
            for i in trials
        ])

        cumulated_len = np.cumsum(
            [trial_change_idx[i + 1] - trial_change_idx[i] for i in trials])
        self.concat_idx = cumulated_len[:-1][np.array(trials[:-1]) +
                                             1 != trials[1:]]


@parametrize(
    "rat-hippocampus-{name}-corrupt-{seed}",
    name=["achilles", "buddy", "cicero", "gatsby"],
    seed=np.arange(1000),
)
class SingleRatCorruptDataset(SingleRatDataset):
    """A single rat hippocampus tetrode recording while the rat navigates on a linear track with a shuffled behavior label.

    Neural data is spike counts binned into 25ms time window and the behavior is position and the running direction (left, right) of a rat.
    The behavior label is structured as 3D array consists of position, right, and left and it is shuffled in random orders.

    Args:
        name: The name of the rat to use. Choose among 'achilles', 'buddy', 'cicero' and 'gatsby'.
        seed: The random seed to set the shuffling.

    """

    def __init__(self, name, seed, root=_DEFAULT_DATADIR):
        super().__init__(name=name, root=root)
        rng = np.random.Generator(np.random.PCG64(seed))
        shuffled_index = np.arange(len(self.index))
        rng.shuffle(shuffled_index)
        self.index = self.index[shuffled_index]


@register("rat-hippocampus-multisubjects-3fold-trial-split")
@parametrize("rat-hippocampus-multisubjects-3fold-trial-split-{split_no}",
             split_no=[0, 1, 2])
class MultipleRatsTrialSplitDataset(cebra.data.DatasetCollection):
    """4 rats hippocampus tetrode recording while the rat navigates on a linear track with 3-fold splits.

    Neural and behavior recordings of 4 rats.
    For each rat, neural data is spike counts binned into 25ms time window and the behavior is position
    and the running direction (left, right) of a rat.
    The behavior label is structured as 3D array consists of position, right, and left.
    Neural and behavior recordings of each rat are parsed into trials (a round trip from one end of the track)
    and the trials are split into a train, valid and test set with k=3 nested cross validation.

    Args:
        split_no: The `k` for k-fold split. Choose among 0, 1, and 2.
        split: The split to use. Choose among 'train', 'valid', 'test', 'all', and 'wo_test'
            (all trials except test split).

    """

    def __init__(self, split_no=0, split=None):
        super().__init__(
            *[
                init(f"rat-hippocampus-{name}-3fold-trial-split-{split_no}",
                     split=split)
                for name in ["achilles", "buddy", "cicero", "gatsby"]
            ],)
        self.names = [dataset.name for dataset in self._datasets]
        self.shapes = [dataset.neural.shape for dataset in self._datasets]
        self._split = split

    def __repr__(self):
        return (
            f"MultipleRatsTrialSplitDataset(name: {self.names}, shape: {self.shapes})"
        )

    def split(self, split):
        assert split == self._split



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for simulating spike counts for Poisson-like neuron models."""

import dataclasses

import joblib
import numpy as np
import scipy.stats
import torch
import torch.distributions
from torch import nn


class PoissonNeuronTransform(nn.Module):
    """Transform spike rates into expected spike counts.

    This is an implementation for transforming arrays or tensors containing spike
    rates into expected spike counts.

    Args:
        num_neurons: The number of neurons to simulate. Needs to match the
            dimensions of the array passed to :py:meth:`__call__`.
        refractory_period: The neuron's absolute refractory period, in seconds.
            The absolute refactory period is the lower bound for the inter-spike
            interval for each neuron.

    References:
        https://neuronaldynamics.epfl.ch/online/Ch7.S3.html
    """

    def __init__(self, num_neurons: int, refractory_period: float = 0.0):
        super().__init__()
        if refractory_period < 0:
            raise ValueError(
                f"Refractory period needs to be non-negative, but got {refractory_period}"
            )
        if num_neurons <= 0:
            raise ValueError(
                f"num_neurons needs to be positive, but got {num_neurons}")
        self.refractory_period = refractory_period
        self.num_neurons = num_neurons

    def __call__(self, spike_rates: torch.Tensor) -> torch.Tensor:
        """Sample spike counts from spike rates

        Args:
            spike_rates: The non-negative spike rates for each neuron, in a
                tensor with shape ``neurons x trials x timesteps``. The number
                of neurons needs to match :py:attr:`num_neurons`.

        Returns:
            A tensor of same shape as the input array, containing a sample
            of spike counts.
        """
        n_neurons, n_trials, n_timesteps = spike_rates.shape
        assert n_neurons == self.num_neurons

        time_interval = 1

        for n_sigmas in [4, 8, 12, 16, 24, 48, 96]:
            num_spikes = int(spike_rates.max() * (time_interval *
                                                  (1 + n_sigmas)))

            # NOTE(stes): See https://pytorch.org/docs/stable/distributions.html#exponential, the
            # parameter of the distribution is directly 1/scale.
            delta_distribution = torch.distributions.Exponential(
                rate=spike_rates)

            # NOTE(stes): The sample dimension is appended to the front of the array (dim 0)
            deltas = delta_distribution.sample(
                (num_spikes,)) + self.refractory_period

            spike_times = torch.cumsum(deltas, dim=0)

            # At least 2 spikes should occur after the window to avoid bias
            if (spike_times[-2] > time_interval).all():
                break
        else:
            raise ValueError()
        return (spike_times < time_interval).sum(dim=0)


def _sample_batch(spike_rates: torch.Tensor, refractory_period: float = 0):
    sample_poisson = PoissonNeuronTransform(num_neurons=spike_rates.shape[0],
                                            refractory_period=refractory_period)
    return sample_poisson(spike_rates)


def sample_parallel(spike_rates,
                    refractory_period: float = 0.0,
                    n_jobs: int = 10):
    """Generate spike counts from the specified spike rates.

    Args:
        spike_rates: The (non-negative) spike rates, with shape ``neurons x trials x time.``
            The number of neurons needs to be divisible by ``n_jobs``.
        n_jobs: The number of parallel jobs for generating the spike trains.

    Returns:
        A tensor of shape ``neurons x trials x time`` which contains the spike counts.
    """

    if any(spike_rates.flatten() < 0):
        raise ValueError("Need to pass non-negative values as the spike rate.")

    spike_rates = spike_rates.view(n_jobs, -1, *spike_rates.shape[1:])
    _sample_batch_delayed = joblib.delayed(_sample_batch)
    spike_counts = joblib.Parallel(n_jobs)(
        _sample_batch_delayed(batch, refractory_period=refractory_period)
        for batch in spike_rates)
    spike_counts = torch.cat(spike_counts, dim=0)

    return spike_counts


@dataclasses.dataclass
class PoissonNeuron:
    """A Poisson neuron model with different sampling methods.

    This is a reference implementation that can be used for testing generation of
    synthetic datasets with Poisson spiking.

    References:
      https://neuronaldynamics.epfl.ch/online/Ch7.S3.html
    """

    # the neuron's spike rate, in Hz
    spike_rate: float = 40.0

    # the length in seconds of the time interval to consider
    # for estimating the spike count. The count
    # will be the total count for the whole duration.
    time_interval: float = 1

    # the number of repeats. If you see irregularities, increase
    # this value.
    num_repeats: int = 10000

    # how many sigmas of the poisson distribution to consider.
    # you can leave this typically in the range of 1-3 sigmas
    data_range_sigmas: float = 10

    @property
    def _expected_count(self):
        """The expected number of spike counts, estimated as lambda * (1 + data_range_sigmas)"""
        return int(self.spike_rate * self.time_interval *
                   (1 + self.data_range_sigmas))

    def _get_counts(self, refractory_period: float = 0.0):
        """Estimate spike counts by sampling individual spikes"""
        max_spike_count = self._expected_count
        deltas = scipy.stats.expon(scale=1.0 / self.spike_rate).rvs(
            (self.num_repeats, max_spike_count))
        deltas += refractory_period
        spike_times = np.cumsum(deltas, axis=-1)

        latest_spike_time = spike_times.max(axis=-1).min()
        if latest_spike_time < self.time_interval:
            raise ValueError(
                f"The simulated number of spikes were not sufficient to complete the simulation. "
                f"You specified a time interval of {self.time_interval}s, but the last spike occurred "
                f"at {latest_spike_time}s. We simulated {max_spike_count} spikes in total. "
                f"Try to either increase data_range_sigmas (={self.data_range_sigmas}) or to decrease "
                f"the time_interval.")

        return (spike_times < self.time_interval).sum(axis=-1)

    def sample_spikes(self, refractory_period: float = 0.0):
        """Estimate count histogram by actually simulating spikes.

        Args:
            refractory_period: The refractory period (= cutoff of the exponential distribution
                we sample the interspike intervals from).

        Returns:
            histogram (bins and counts) of spike counts, estimated over the specified
            number of repeats.
        """
        total_counts = self._get_counts(refractory_period=refractory_period)
        histogram = np.bincount(total_counts)
        return np.arange(len(histogram)), histogram

    def sample_poisson(self, spike_rate=None, range_=None):
        """Compute count histogram with a homogeneous Poisson process.

        Returns:
            histogram (bins and counts) of spike counts, estimated over the specified
            number of repeats.
        """
        if spike_rate is None:
            spike_rate = self.spike_rate
        if range_ is None:
            range_ = (0, self._expected_count)
        t = np.arange(*range_)
        poisson = scipy.stats.poisson(spike_rate * self.time_interval).pmf
        return t, np.round(poisson(t) * self.num_repeats)

    def sample_poisson_estimate(self,
                                refractory_period: float = 0.0,
                                range_=None):
        """Compute count histogram with a Poisson distribution fitted to simulated data.

        The function first samples spikes using :py:meth:`sample_spikes`, then estimates the
        maximum likelihood fit of a Poisson distribution (by computing the mean number of spikes
        over all repeats).

        The count histogram is then computed from this estimated poisson distribution.

        Returns:
            histogram (bins and counts) of spike counts, estimated over the specified
            number of repeats.
        """
        total_counts = self._get_counts(
            refractory_period=refractory_period).mean()
        return self.sample_poisson(spike_rate=total_counts / self.time_interval,
                                   range_=range_)



#
# CEBRA: Consistent EmBeddings of high-dimensional Recordings using Auxiliary variables
# © Mackenzie W. Mathis & Steffen Schneider (v0.4.0+)
# Source code:
# https://github.com/AdaptiveMotorControlLab/CEBRA
#
# Please see LICENSE.md for the full license document:
# https://github.com/AdaptiveMotorControlLab/CEBRA/blob/main/LICENSE.md
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integrations tests using various usecases via the sklearn API.

All attempts to reproduce bugs in the sklearn API should be added to this
file. Existing tests should **not** be changed to ensure backward compatibility.

If breaking changes are introduced, add decorators that skip tests based on the
cebra version we test against.
"""

import itertools

import numpy as np
import pytest
import torch

import cebra
import cebra.integrations.sklearn.dataset as cebra_sklearn_dataset


def _default_kwargs():
    return dict(device="cpu", max_iterations=10, verbose=False)


def _make_data():
    spikes = np.random.normal(0, 1, size=(1000, 10))
    behavior = np.random.normal(0, 1, size=(1000, 5))
    return spikes, behavior


def _make_data_torch():
    spikes = torch.randn(size=(1000, 10))
    behavior = torch.randn(size=(1000, 5))
    return spikes, behavior


def _run_test(model, behavior=False):
    neural, index = _make_data()
    print(index)
    if behavior:
        model.fit(neural, index)
    else:
        model.fit(neural)
    embedding = model.transform(neural)
    assert len(embedding) == len(neural)
    assert len(embedding.shape) == 2


def test_minimal():
    model = cebra.CEBRA("offset10-model",
                        output_dimension=2,
                        time_offsets=10,
                        **_default_kwargs())
    _run_test(model)
    _run_test(model, behavior=True)


def test_full():
    model = cebra.CEBRA("offset10-model",
                        output_dimension=3,
                        time_offsets=(10,),
                        conditional="time_delta",
                        batch_size=300,
                        **_default_kwargs())
    _run_test(model)
    _run_test(model, behavior=True)


def test_hybrid():
    model = cebra.CEBRA("offset10-model",
                        output_dimension=3,
                        time_offsets=(10,),
                        conditional="time_delta",
                        batch_size=300,
                        hybrid=True,
                        **_default_kwargs())
    _run_test(model, behavior=True)
    with pytest.raises(ValueError):
        _run_test(model, behavior=False)


_args = [
    dict(
        model_architecture="offset1-model",
        time_offsets=10,
        batch_size=200,
        learning_rate=3e-4,
        num_hidden_units=64,
        output_dimension=8,
        temperature=1,
    ),
    dict(
        model_architecture="offset1-model",
        batch_size=200,
        learning_rate=3e-4,
        num_hidden_units=64,
        temperature=1,
    ),
    dict(model_architecture="offset1-model", time_offsets=10),
]
_additions = [
    dict(),
]
_args = list(
    dict(**arg, **addition)
    for arg, addition in itertools.product(_args, _additions))


@pytest.mark.parametrize(
    "leave_out,args",
    [(leave_out, args) for args in _args for leave_out in args.keys()])
def test_leave_arg_out(leave_out, args):
    model = cebra.CEBRA(**{
        k: v for k, v in args.items() if k != leave_out
    }, **_default_kwargs())
    _run_test(model)


def test_defaults():
    model = cebra.CEBRA(max_iterations=10)
    _run_test(model)


def test_dataset():
    X = np.zeros((100, 3))
    y = np.zeros((100, 3))
    cebra_sklearn_dataset.SklearnDataset(X, y=())
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X, y=None)
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X, (y,), device="foo")
    with pytest.raises(ValueError):
        cebra_sklearn_dataset.SklearnDataset(X=None, y=(y,))

    dataset = cebra_sklearn_dataset.SklearnDataset(X, y=(y,))
    assert dataset.input_dimension == X.shape[1]
    assert torch.allclose(dataset.continuous_index, torch.from_numpy(y).float())
    assert dataset.continuous_index_dimensions == y.shape[1]
    assert dataset.total_index_dimensions == y.shape[1]


def test_incompatible():
    X, y = _make_data()

    # hybrid training is not possible without passing
    # and index
    model = cebra.CEBRA(hybrid=True)
    with pytest.raises(ValueError):
        model.fit(X)

    # multisession training is not possible without passing
    # an index
    model = cebra.CEBRA()
    with pytest.raises(RuntimeError):
        model.fit([X, X])


# @pytest.mark.timeout(3)
# def test_no_cuda():
#    # Then initializing a new model can be done by running:
#    #if torch.cuda.is_available():
#    #    pytest.skip("Test only useful when running on CPU.")
#    model = cebra.CEBRA(
#        "offset1-model",
#        **_default_kwargs()
#    )
#    _run_test(model)
